#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.28 09:00:00                  #
# ================================================== #

from typing import Tuple, List
from functools import lru_cache

import tiktoken

# from langchain_core.messages import ChatMessage as ChatMessageLangchain
from llama_index.core.base.llms.types import ChatMessage as ChatMessageLlama

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_ASSISTANT,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_RESEARCH,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_AGENT_OPENAI,
    MODE_COMPUTER,
)
from pygpt_net.item.ctx import CtxItem

CHAT_MODES = [
    MODE_CHAT,
    # MODE_VISION,
    # MODE_LANGCHAIN,
    MODE_ASSISTANT,
    MODE_LLAMA_INDEX,
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_AGENT_OPENAI,
    MODE_EXPERT,
    MODE_AUDIO,
    MODE_RESEARCH,
    MODE_COMPUTER,
]

class Tokens:
    _default_encoding = "cl100k_base"
    _const_cache = {}

    def __init__(self, window=None):
        self.window = window

    @staticmethod
    @lru_cache(maxsize=128)
    def _encoding_name_for_model(model: str | None) -> str:
        if model:
            try:
                return tiktoken.encoding_for_model(model).name
            except KeyError:
                return Tokens._default_encoding
            except ValueError:
                return Tokens._default_encoding
        return Tokens._default_encoding

    @staticmethod
    @lru_cache(maxsize=64)
    def _get_encoding(encoding_name: str):
        try:
            return tiktoken.get_encoding(encoding_name)
        except Exception:
            return tiktoken.get_encoding(Tokens._default_encoding)

    @classmethod
    def _const_tokens(cls, text: str, model: str = "gpt-4") -> int:
        enc_name = cls._encoding_name_for_model(model)
        key = (enc_name, text)
        cached = cls._const_cache.get(key)
        if cached is not None:
            return cached
        try:
            enc = cls._get_encoding(enc_name)
            val = len(enc.encode(text))
        except Exception:
            val = 0
        cls._const_cache[key] = val
        return val

    @staticmethod
    def from_str(
            string: str,
            model: str = "gpt-4"
    ) -> int:
        if not string:
            return 0
        try:
            try:
                enc_name = Tokens._encoding_name_for_model(model)
                encoding = Tokens._get_encoding(enc_name)
            except ValueError:
                return 0
            try:
                return len(encoding.encode(str(string)))
            except Exception as e:
                print("Tokens calc exception", e)
                return 0
        except Exception as e:
            print("Tokens calculation exception:", e)
            return 0

    @staticmethod
    def get_extra(model: str = "gpt-4") -> int:
        return 3

    @staticmethod
    def from_prompt(
            text: str,
            name: str,
            model: str = "gpt-4"
    ) -> int:
        model, per_message, per_name = Tokens.get_config(model)
        num = 0
        try:
            num += Tokens.from_str(text, model)
        except Exception as e:
            print("Tokens calc exception", e)
        if name:
            num += per_message + per_name
        else:
            num += per_message
        return num

    @staticmethod
    def from_text(
            text: str,
            model: str = "gpt-4"
    ) -> int:
        if not text:
            return 0
        try:
            return Tokens.from_str(text, model)
        except Exception as e:
            print("Tokens calc exception", e)
            return 0

    @staticmethod
    def from_messages(
            messages: List[dict],
            model: str = "gpt-4"
    ) -> int:
        model, per_message, per_name = Tokens.get_config(model)
        num = 0
        f = Tokens.from_str
        for message in messages:
            num += per_message
            for key, value in message.items():
                if isinstance(value, str):
                    num += f(value)
                    if key == "name":
                        num += per_name
                elif key == "content" and isinstance(value, list):
                    for part in value:
                        t = part.get("type")
                        if t == "text":
                            tv = part.get("text")
                            if tv:
                                num += f(tv)
        num += 3
        return num

    @staticmethod
    def from_langchain_messages(
            messages: List,
            model: str = "gpt-4"
    ) -> int:
        model, per_message, per_name = Tokens.get_config(model)
        num = 0
        f = Tokens.from_str
        for message in messages:
            num += per_message
            num += f(message.content)
        num += 3
        return num

    @staticmethod
    def from_llama_messages(
            query: str,
            messages: List[ChatMessageLlama],
            model: str = "gpt-4"
    ) -> int:
        model, per_message, per_name = Tokens.get_config(model)
        num = 0
        f = Tokens.from_str
        num += f(query)
        for message in messages:
            num += per_message
            num += f(message.content)
        num += 3
        return num

    @staticmethod
    def from_ctx(
            ctx: CtxItem,
            mode: str = MODE_CHAT,
            model: str = "gpt-4"
    ) -> int:
        model, per_message, per_name = Tokens.get_config(model)
        num = 0
        f = Tokens.from_str
        if mode in CHAT_MODES:
            try:
                num += f(str(ctx.final_input), model)
            except Exception as e:
                print("Tokens calc exception", e)
            try:
                num += f(str(ctx.final_output), model)
            except Exception as e:
                print("Tokens calc exception", e)
            num += per_message * 2
            num += per_name * 2
            try:
                num += Tokens._const_tokens("system", model) * 2
            except Exception as e:
                print("Tokens calc exception", e)
            name = ctx.input_name if ctx.input_name else "user"
            try:
                num += f(name, model)
            except Exception as e:
                print("Tokens calc exception", e)
            name = ctx.output_name if ctx.output_name else "assistant"
            try:
                num += f(name, model)
            except Exception as e:
                print("Tokens calc exception", e)
        elif mode == MODE_COMPLETION:
            parts = []
            if ctx.input_name and ctx.output_name:
                if ctx.final_input:
                    parts.append("\n" + ctx.input_name + ": " + ctx.final_input)
                if ctx.final_output:
                    parts.append("\n" + ctx.output_name + ": " + ctx.final_output)
            else:
                if ctx.final_input:
                    parts.append("\n" + ctx.final_input)
                if ctx.final_output:
                    parts.append("\n" + ctx.final_output)
            try:
                num += f("".join(parts), model)
            except Exception as e:
                print("Tokens calc exception", e)
        return num

    def get_current(
            self,
            input_prompt: str
    ) -> Tuple[int, int, int, int, int, int, int, int, int]:
        model = self.window.core.config.get('model')
        model_id = ""
        model_data = self.window.core.models.get(model)
        if model_data is not None:
            model_id = model_data.id
        mode = self.window.core.config.get('mode')
        user_name = self.window.core.config.get('user_name')
        ai_name = self.window.core.config.get('ai_name')

        system_prompt = ""
        system_tokens = 0
        input_tokens = 0
        max_total_tokens = self.window.core.config.get('max_total_tokens')
        extra_tokens = self.get_extra(model)

        if mode in CHAT_MODES:
            system_prompt = str(self.window.core.config.get('prompt')).strip()
            system_prompt = self.window.core.prompt.build_final_system_prompt(system_prompt, mode, model_data)
            if system_prompt:
                system_tokens = self.from_prompt(system_prompt, "", model_id)
                system_tokens += Tokens._const_tokens("system", model_id)
            if input_prompt:
                input_tokens = self.from_prompt(input_prompt, "", model_id)
                input_tokens += Tokens._const_tokens("user", model_id)
        elif mode == MODE_COMPLETION:
            system_prompt = str(self.window.core.config.get('prompt')).strip()
            system_prompt = self.window.core.prompt.build_final_system_prompt(system_prompt, mode, model_data)
            system_tokens = self.from_text(system_prompt, model_id)
            if input_prompt:
                if user_name and ai_name:
                    message = "\n" + user_name + ": " + str(input_prompt) + "\n" + ai_name + ":"
                else:
                    message = "\n" + str(input_prompt)
                input_tokens = self.from_text(message, model_id)
                extra_tokens = 0

        self.window.core.ctx.current_sys_prompt = system_prompt

        used_tokens = system_tokens + input_tokens

        max_current = max_total_tokens
        model_ctx = self.window.core.models.get_num_ctx(model)
        if max_current > model_ctx:
            max_current = model_ctx

        threshold = self.window.core.config.get('context_threshold')
        max_to_check = max_current - threshold

        ctx_len_all = self.window.core.ctx.count_items()
        ctx_len, ctx_tokens = self.window.core.ctx.count_prompt_items(model_id, mode, used_tokens, max_to_check)

        if not self.window.core.config.get('use_context'):
            ctx_tokens = 0
            ctx_len = 0

        sum_tokens = system_tokens + input_tokens + ctx_tokens + extra_tokens

        return input_tokens, system_tokens, extra_tokens, ctx_tokens, ctx_len, ctx_len_all, \
               sum_tokens, max_current, threshold

    def from_user(
            self,
            system_prompt: str,
            input_prompt: str
    ) -> int:
        model = self.window.core.config.get('model')
        model_id = self.window.core.models.get_id(model)
        mode = self.window.core.config.get('mode')
        tokens = 0
        if mode in [MODE_CHAT, MODE_AUDIO, MODE_RESEARCH]:
            tokens += self.from_prompt(system_prompt, "", model_id)
            tokens += self.from_text("system", model_id)
            tokens += self.from_prompt(input_prompt, "", model_id)
            tokens += self.from_text("user", model_id)
        else:
            tokens += self.from_text(system_prompt, model_id)
            tokens += self.from_text(input_prompt, model_id)
        tokens += self.window.core.config.get('context_threshold')
        tokens += self.get_extra(model_id)
        return tokens

    @staticmethod
    def get_config(model: str) -> Tuple[str, int, int]:
        per_message = 4
        per_name = -1

        if model is not None:
            if model in {
                "gpt-3.5-turbo-0613",
                "gpt-3.5-turbo-16k-0613",
                "gpt-4-0314",
                "gpt-4-32k-0314",
                "gpt-4-0613",
                "gpt-4-32k-0613",
            }:
                per_message = 3
                per_name = 1
            elif model == "gpt-3.5-turbo-0301":
                per_message = 4
                per_name = -1
            elif "gpt-3.5-turbo" in model:
                return Tokens.get_config(model="gpt-3.5-turbo-0613")
            elif "gpt-4" in model:
                return Tokens.get_config(model="gpt-4-0613")
            elif model.startswith("text-davinci"):
                per_message = 1
                per_name = 0

        return model, per_message, per_name


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.26 13:00:00                  #
# ================================================== #

import copy
from typing import Dict, List, Optional

from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItemModel, QIcon
from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QScrollArea, QWidget, QTabWidget, QFrame, \
    QSplitter, QSizePolicy

from pygpt_net.ui.widget.dialog.model import ModelDialog
from pygpt_net.ui.widget.element.group import CollapsedGroup
from pygpt_net.ui.widget.element.labels import UrlLabel, DescLabel
from pygpt_net.ui.widget.lists.model_editor import ModelEditorList
from pygpt_net.ui.widget.option.checkbox import OptionCheckbox
from pygpt_net.ui.widget.option.checkbox_list import OptionCheckboxList
from pygpt_net.ui.widget.option.combo import OptionCombo
from pygpt_net.ui.widget.option.dictionary import OptionDict
from pygpt_net.ui.widget.option.input import OptionInput, PasswordInput
from pygpt_net.ui.widget.option.slider import OptionSlider
from pygpt_net.ui.widget.option.textarea import OptionTextarea
from pygpt_net.ui.widget.textarea.search_input import SearchInput
from pygpt_net.utils import trans


class Models:
    def __init__(self, window=None):
        """
        Models editor dialog

        :param window: Window instance
        """
        self.window = window
        self.dialog_id = "models.editor"

        # Internal state for filtering/mapping
        self._filter_text: str = ""
        self._filtered_ids: List[str] = []     # current list of model keys displayed in the view (order matters)
        self._index_to_id: List[str] = []      # row index -> model key
        self._id_to_index: Dict[str, int] = {} # model key -> row index
        self._all_data: Dict[str, object] = {} # last known (unfiltered) data snapshot used to render the list

    def setup(self, idx=None):
        """
        Setup editor dialog

        :param idx: current model tab index
        """
        self.window.ui.nodes['models.editor.btn.new'] = \
            QPushButton(QIcon(":/icons/add.svg"), trans("dialog.models.editor.btn.new"))
        self.window.ui.nodes['models.editor.btn.defaults.user'] = \
            QPushButton(trans("dialog.models.editor.btn.defaults.user"))
        self.window.ui.nodes['models.editor.btn.defaults.app'] = \
            QPushButton(trans("dialog.models.editor.btn.defaults.app"))
        self.window.ui.nodes['models.editor.btn.save'] = \
            QPushButton(trans("dialog.models.editor.btn.save"))

        self.window.ui.nodes['models.editor.btn.new'].clicked.connect(
            lambda: self.window.controller.model.editor.new()
        )
        self.window.ui.nodes['models.editor.btn.defaults.user'].clicked.connect(
            lambda: self.window.controller.model.editor.load_defaults_user()
        )
        self.window.ui.nodes['models.editor.btn.defaults.app'].clicked.connect(
            lambda: self.window.controller.model.editor.load_defaults_app()
        )
        self.window.ui.nodes['models.editor.btn.save'].clicked.connect(
            lambda: self.window.controller.model.editor.save()
        )

        # set enter key to save button
        self.window.ui.nodes['models.editor.btn.new'].setAutoDefault(False)
        self.window.ui.nodes['models.editor.btn.defaults.user'].setAutoDefault(False)
        self.window.ui.nodes['models.editor.btn.defaults.app'].setAutoDefault(False)
        self.window.ui.nodes['models.editor.btn.save'].setAutoDefault(True)

        # footer buttons
        footer = QHBoxLayout()
        footer.addWidget(self.window.ui.nodes['models.editor.btn.new'])
        footer.addWidget(self.window.ui.nodes['models.editor.btn.defaults.user'])
        footer.addWidget(self.window.ui.nodes['models.editor.btn.defaults.app'])
        footer.addWidget(self.window.ui.nodes['models.editor.btn.save'])

        # editor tabs
        self.window.ui.tabs['models.editor'] = QTabWidget()

        # build settings tabs
        parent_id = "model"

        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        content = QVBoxLayout()

        # create config options entry if not exists
        if parent_id not in self.window.ui.config:
            self.window.ui.config[parent_id] = {}

        # get options
        options = copy.deepcopy(self.window.controller.model.editor.get_options())
        widgets = self.build_widgets(options)
        advanced_keys = []
        for key in options:
            if 'advanced' in options[key] and options[key]['advanced']:
                advanced_keys.append(key)

        # apply settings widgets
        for key in widgets:
            self.window.ui.config[parent_id][key] = widgets[key]

        for key in widgets:
            if key in advanced_keys:  # hide advanced options
                continue
            content.addLayout(self.add_option(widgets[key], options[key]))  # add to scroll

        # append advanced options at the end
        if len(advanced_keys) > 0:
            group_id = 'models.editor.advanced'
            group = CollapsedGroup(self.window, group_id, None, False, None)
            group.box.setText(trans('settings.advanced.collapse'))
            for key in widgets:
                if key not in advanced_keys:  # ignore non-advanced options
                    continue
                group.add_layout(self.add_option(widgets[key], options[key]))  # add option to group

            # add advanced options group to scroll
            content.addWidget(group)
            self.window.ui.groups[group_id] = group

        content.addStretch()

        # scroll widget
        scroll_widget = QWidget()
        scroll_widget.setLayout(content)
        scroll.setWidget(scroll_widget)

        area = QVBoxLayout()
        area.addWidget(scroll)

        area_widget = QWidget()
        area_widget.setLayout(area)

        data = {}
        for model_id in self.window.core.models.items.keys():
            model = self.window.core.models.items[model_id]
            data[model_id] = model

        # models list
        id = 'models.list'
        self.window.ui.nodes[id] = ModelEditorList(self.window, id)
        self.window.ui.models[id] = self.create_model(self.window)
        self.window.ui.nodes[id].setModel(self.window.ui.models[id])
        self.window.ui.nodes[id].setMinimumWidth(250)  # set max width to list

        # search input (placed above the list)
        self.window.ui.nodes['models.editor.search'] = SearchInput()
        # Connect to provided callback API (callables assigned to attributes)
        self.window.ui.nodes['models.editor.search'].on_search = self._on_search_models
        self.window.ui.nodes['models.editor.search'].on_clear = self._on_clear_models  # clear via "X" button

        # provider select
        option_provider = self.window.controller.model.editor.get_provider_option()
        self.window.ui.config[parent_id]['provider_global'] = OptionCombo(self.window, parent_id, 'provider_global',
                                                                          option_provider)
        provider_keys = self.window.controller.config.placeholder.apply_by_id('llm_providers')
        provider_keys.insert(0, {"-": trans("list.all")})  # add "All" option
        self.window.ui.config[parent_id]['provider_global'].set_keys(provider_keys)

        # container for search + list (left panel)
        left_layout = QVBoxLayout()
        left_layout.setContentsMargins(0, 0, 0, 0)
        left_layout.setSpacing(6)
        left_layout.addWidget(self.window.ui.nodes['models.editor.search'])
        left_layout.addWidget(self.window.ui.config[parent_id]['provider_global'])
        left_layout.addWidget(self.window.ui.nodes[id])
        left_widget = QWidget()
        left_widget.setLayout(left_layout)

        # update models list (initial, unfiltered)
        self.update_list(id, data)

        # splitter
        self.window.ui.splitters['dialog.models'] = QSplitter(Qt.Horizontal)
        self.window.ui.splitters['dialog.models'].addWidget(left_widget)  # search + list
        self.window.ui.splitters['dialog.models'].addWidget(area_widget)  # tabs
        self.window.ui.splitters['dialog.models'].setStretchFactor(0, 2)
        self.window.ui.splitters['dialog.models'].setStretchFactor(1, 5)
        self.window.ui.splitters['dialog.models'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        main_layout = QHBoxLayout()
        main_layout.addWidget(self.window.ui.splitters['dialog.models'])

        layout = QVBoxLayout()
        layout.addLayout(main_layout)  # list + model tabs
        layout.addLayout(footer)  # bottom buttons (save, defaults)

        self.window.ui.dialog[self.dialog_id] = ModelDialog(self.window, self.dialog_id)
        self.window.ui.dialog[self.dialog_id].setLayout(layout)
        self.window.ui.dialog[self.dialog_id].setWindowTitle(trans('dialog.models.editor'))

        # restore current opened tab if idx is set
        if idx is not None:
            try:
                self.window.controller.model.editor.set_by_tab(idx)
            except Exception as e:
                print('Failed restore models editor tab: {}'.format(idx))
        else:
            if self.window.controller.model.editor.current is None:
                self.window.controller.model.editor.set_by_tab(0)

    def _on_search_models(self, text: str):
        """
        Handle SearchInput callback. Filtering is prefix-based on model id or name (case-insensitive).
        """
        # Keep normalized filter text
        self._filter_text = (text or "").strip().casefold()
        # Re-apply list render using the last known dataset
        self.update_list('models.list', self._all_data)
        # Do not force-select rows here to avoid heavy editor re-initialization on each keystroke.

    def _on_clear_models(self):
        """
        Handle SearchInput clear (click on 'X' button).
        """
        if not self._filter_text:
            return
        self._filter_text = ""
        self.update_list('models.list', self._all_data)
        # Try to restore selection for the current model if it is visible again.
        self._restore_selection_for_current()

    def _restore_selection_for_current(self):
        """
        Attempt to restore selection on the list to the current editor model if present in the view.
        """
        current_id = getattr(self.window.controller.model.editor, "current", None)
        if not current_id:
            return
        idx = self.get_row_by_model_id(current_id)
        if idx is None:
            return
        current = self.window.ui.models['models.list'].index(idx, 0)
        self.window.ui.nodes['models.list'].setCurrentIndex(current)

    def _apply_filter(self, data: Dict[str, object]) -> Dict[str, object]:
        """
        Return filtered data view, preserving original order.
        """
        if not self._filter_text:
            return data

        out: Dict[str, object] = {}
        needle = self._filter_text
        for mid, item in data.items():
            # Normalize fields for prefix matching
            model_id = (str(getattr(item, "id", mid)) or "").casefold()
            model_name = (getattr(item, "name", "") or "").casefold()
            if (needle in model_name
                    or needle in model_id):
                out[mid] = item
        return out

    def _refresh_index_mapping(self, ids: List[str]):
        """
        Build fast row<->id mapping for the current filtered list.
        """
        self._filtered_ids = list(ids)
        self._index_to_id = list(ids)
        self._id_to_index = {mid: idx for idx, mid in enumerate(ids)}

    def get_model_id_by_row(self, row: int) -> Optional[str]:
        """
        Map a view row index to the model id currently displayed at that row.
        """
        if 0 <= row < len(self._index_to_id):
            return self._index_to_id[row]
        return None

    def get_row_by_model_id(self, model_id: str) -> Optional[int]:
        """
        Map a model id to its current row index in the filtered view.
        """
        return self._id_to_index.get(model_id)

    def get_filtered_ids(self) -> List[str]:
        """
        Return a copy of currently visible model ids (filtered order).
        """
        return list(self._filtered_ids)

    def build_widgets(self, options: dict) -> dict:
        """
        Build settings options widgets

        :param options: model options
        :return: dict of widgets
        """
        parent = "model"
        widgets = {}

        for key in options:
            option = options[key]
            option['id'] = key
            # create widget by option type
            if option['type'] == 'text' or option['type'] == 'int' or option['type'] == 'float':
                if 'slider' in option and option['slider'] \
                        and (option['type'] == 'int' or option['type'] == 'float'):
                    widgets[key] = OptionSlider(self.window, parent, key, option)  # slider + text input
                else:
                    if 'secret' in option and option['secret']:
                        widgets[key] = PasswordInput(self.window, parent, key, option)  # password input
                    else:
                        widgets[key] = OptionInput(self.window, parent, key, option)  # text input
            elif option['type'] == 'textarea':
                widgets[key] = OptionTextarea(self.window, parent, key, option)  # textarea
            elif option['type'] == 'bool':
                widgets[key] = OptionCheckbox(self.window, parent, key, option)  # checkbox
            elif option['type'] == 'bool_list':
                self.window.controller.config.placeholder.apply(option)
                widgets[key] = OptionCheckboxList(self.window, parent, key, option)  # checkbox list
            elif option['type'] == 'dict':
                self.window.controller.config.placeholder.apply(option)
                widgets[key] = OptionDict(self.window, parent, key, option)  # dictionary
                widgets[key].setMinimumHeight(200)
                # register dict to editor:
                self.window.ui.dialogs.register_dictionary(
                    key,
                    parent=parent,
                    option=option,
                )
            elif option['type'] == 'combo':
                self.window.controller.config.placeholder.apply(option)
                widgets[key] = OptionCombo(self.window, parent, key, option)  # combobox

        return widgets

    def add_line(self) -> QFrame:
        """
        Make separator line

        :return: separator line
        """
        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        return line

    def add_urls(self, urls: dict) -> QWidget:
        """
        Add clickable urls to list

        :param urls: urls dict
        :return: QWidget
        """
        layout = QVBoxLayout()
        for name in urls:
            url = urls[name]
            label = UrlLabel(name, url)
            layout.addWidget(label)

        widget = QWidget()
        widget.setLayout(layout)
        return widget

    def add_option(self, widget: QWidget, option: dict) -> QVBoxLayout:
        """
        Append option widget to layout

        :param widget: widget instance
        :param option: option dict
        :return: QVBoxLayout
        """
        one_column_types = ['textarea', 'dict', 'bool']
        key = option['id']
        label = trans(option['label'])
        label_key = 'model.' + key + '.label'
        desc = None
        desc_key = None

        if option['type'] != 'bool':
            self.window.ui.nodes[label_key] = QLabel(label)
            self.window.ui.nodes[label_key].setStyleSheet("font-weight: bold;")

        if "description" in option:
            desc = trans(option['description'])
            desc_key = 'model.' + key + '.desc'

        # 2-columns layout
        if option['type'] not in one_column_types:
            cols = QHBoxLayout()
            cols.addWidget(self.window.ui.nodes[label_key], 1)  # disable label in bool type
            cols.addWidget(widget, 4)

            cols_widget = QWidget()
            cols_widget.setLayout(cols)

            if option['type'] != 'bool_list':
                cols_widget.setMaximumHeight(90)

            layout = QVBoxLayout()
            layout.addWidget(cols_widget)
        else:
            # 1-column layout: textarea and dict fields
            layout = QVBoxLayout()
            if option['type'] != 'bool':
                layout.addWidget(self.window.ui.nodes[label_key])
            else:
                widget.box.setText(label)  # set checkbox label
            layout.addWidget(widget)

        if desc:
            self.window.ui.nodes[desc_key] = DescLabel(desc)
            self.window.ui.nodes[desc_key].setMaximumHeight(40)
            layout.addWidget(self.window.ui.nodes[desc_key])

        line = self.add_line()  # TODO: change name to separator
        layout.addWidget(line)

        return layout

    def add_raw_option(self, option: dict) -> QHBoxLayout:
        """
        Add raw option row

        :param option: Option
        :return: QHBoxLayout
        """
        layout = QHBoxLayout()
        layout.addWidget(option)
        return layout

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model
        :param parent: parent widget
        :return: QStandardItemModel
        """
        return QStandardItemModel(0, 1, parent)

    def _get_provider_display_name(self, provider_id: str) -> str:
        """
        Resolve provider display name using app's LLM provider registry.
        """
        if not provider_id:
            return ""
        try:
            name = self.window.core.llm.get_provider_name(provider_id)
            if isinstance(name, str):
                name = name.strip()
            return name or ""
        except Exception:
            # Fail silently to avoid breaking the models list rendering
            return ""

    def update_list(self, id: str, data: dict):
        """
        Update list

        :param id: ID of the list
        :param data: Data to update
        """
        # Keep latest source data for subsequent filtering cycles
        self._all_data = dict(data)

        # Apply current filter (preserve insertion/sorted order)
        filtered = self._apply_filter(self._all_data)

        # Reset model rows
        self.window.ui.models[id].removeRows(0, self.window.ui.models[id].rowCount())

        # Count occurrences of each model display name to detect duplicates (in filtered view)
        name_counts = {}
        for key in filtered:
            item = filtered[key]
            base_name = getattr(item, "name", "") or ""
            name_counts[base_name] = name_counts.get(base_name, 0) + 1

        # Populate rows with optional provider suffix for duplicate names
        i = 0
        row_ids: List[str] = []
        for n in filtered:
            item = filtered[n]
            base_name = getattr(item, "name", "") or ""
            display_name = base_name
            if name_counts.get(base_name, 0) > 1:
                provider_id = getattr(item, "provider", None)
                provider_name = self._get_provider_display_name(provider_id) if provider_id else ""
                if provider_name:
                    display_name = f"{base_name} ({provider_name})"

            self.window.ui.models[id].insertRow(i)
            self.window.ui.models[id].setData(self.window.ui.models[id].index(i, 0), display_name)
            # IMPORTANT: map list row to the dictionary key (stable and unique within models.items)
            row_ids.append(n)
            i += 1

        # Refresh index mappings used by Editor
        self._refresh_index_mapping(row_ids)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 23:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction

from pygpt_net.utils import trans


class Video:
    def __init__(self, window=None):
        """
        Menu setup

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Setup video menu"""
        w = self.window
        ui_menu = w.ui.menu
        cam = w.controller.camera

        capture = QAction(trans("menu.video.capture"), w, checkable=True)
        capture.setToolTip(trans('vision.capture.enable.tooltip'))
        capture.triggered.connect(cam.toggle)

        capture_auto = QAction(trans("menu.video.capture.auto"), w, checkable=True)
        capture_auto.setToolTip(trans('vision.capture.auto.tooltip'))
        capture_auto.triggered.connect(cam.toggle_auto)

        menu_video = w.menuBar().addMenu(trans("menu.video"))
        menu_video.addActions([capture, capture_auto])
        menu_video.setToolTipsVisible(True)

        ui_menu['video.capture'] = capture
        ui_menu['video.capture.auto'] = capture_auto
        ui_menu['menu.video'] = menu_video


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.25 14:00:00                  #
# ================================================== #

import copy
from uuid import uuid4

from pygpt_net.core.types import MODEL_DEFAULT
from pygpt_net.item.agent import AgentItem
from pygpt_net.item.builder_layout import BuilderLayoutItem
from pygpt_net.provider.core.agent.json_file import JsonFileProvider
from pygpt_net.utils import trans


class Custom:

    CUSTOM_AGENT_SUFFIX = " *"  # suffix for custom agents in lists

    def __init__(self, window=None):
        """
        Custom agents core

        :param window: Window instance
        """
        self.window = window
        self.provider = JsonFileProvider(window)  # JSON file provider
        self.agents = {}  # dict of AgentItem
        self.layout = None  # BuilderLayoutItem
        self.loaded = False

    def load(self):
        """Load custom agents from provider"""
        data = self.provider.load()
        if "layout" in data:
            self.layout = data["layout"]
        if "agents" in data:
            self.agents = data["agents"]
        self.loaded = True

    def reload(self):
        """Reload custom agents from provider"""
        self.loaded = False
        self.load()

    def save(self):
        """Save custom agents to provider"""
        self.provider.save(self.layout, self.agents)

    def is_custom(self, agent_id: str) -> bool:
        """
        Check if agent is custom

        :param agent_id: agent ID
        :return: True if custom
        """
        if not self.loaded:
            self.load()
        return agent_id in self.agents

    def update_layout(self, layout: dict):
        """
        Update current layout of custom agents editor

        :param layout: layout dict
        """
        if self.layout is None:
            self.layout = BuilderLayoutItem()
        self.layout.data = layout

    def reset(self):
        """Reset custom agents"""
        self.agents = {}
        self.layout = None
        self.loaded = False
        self.provider.truncate()

    def get_layout(self) -> dict:
        """
        Get layout of custom agents

        :return: layout dict
        """
        if not self.loaded:
            self.load()
        return self.layout

    def get_agents(self) -> dict:
        """
        Get custom agents

        :return: dict of AgentItem
        """
        if not self.loaded:
            self.load()
        return self.agents

    def get_choices(self) -> list:
        """
        Get custom agents choices

        :return: list of dict with 'id' and 'name'
        """
        if not self.loaded:
            self.load()
        return [{agent.id: f"{agent.name}{self.CUSTOM_AGENT_SUFFIX}"} for agent in self.agents.values()]

    def get_agent(self, agent_id: str):
        """
        Get custom agent by ID

        :param agent_id: agent ID
        :return: AgentItem or None
        """
        if not self.loaded:
            self.load()
        return self.agents.get(agent_id)

    def get_ids(self) -> list:
        """
        Get list of custom agent IDs

        :return: list of agent IDs
        """
        if not self.loaded:
            self.load()
        return list(self.agents.keys())

    def get_schema(self, agent_id: str) -> list:
        """
        Get schema of a specific custom agent

        :param agent_id: agent ID
        :return: list with schema or empty list
        """
        if not self.loaded:
            self.load()
        agent = self.agents.get(agent_id)
        if agent:
            return agent.schema
        return []

    def build_options(self, agent_id: str) -> dict:
        """
        Build options for a specific custom agent

        :param agent_id: agent ID
        :return: dict with options or empty dict
        """
        if not self.loaded:
            self.load()

        agent = self.agents.get(agent_id)
        if not agent:
            return {}

        schema = agent.schema if agent else []
        options = {}
        for node in schema:
            try:
                if "type" in node and node["type"] == "agent":
                    if "id" in node:
                        sub_agent_id = node["id"]
                        tab = {
                            "label": sub_agent_id
                        }
                        opts = {
                            "model": {
                                "label": trans("agent.option.model"),
                                "type": "combo",
                                "use": "models",
                                "default": MODEL_DEFAULT,
                            }
                        }
                        if "slots" in node:
                            slots = node["slots"]
                            if "name" in slots and slots["name"]:
                                tab["label"] = slots["name"]
                            if "role" in slots:
                                opts["role"] = {
                                    "type": "text",
                                    "label": trans("agent.option.role.label"),
                                    "description": trans("agent.option.role"),
                                    "default": slots["role"],
                                }
                            if "instruction" in slots:
                                opts["prompt"] = {
                                    "type": "textarea",
                                    "label": trans("agent.option.prompt"),
                                    "default": slots["instruction"],
                                }
                            if "remote_tools" in slots:
                                opts["allow_remote_tools"] = {
                                    "type": "bool",
                                    "label": trans("agent.option.tools.remote"),
                                    "description": trans("agent.option.tools.remote.desc"),
                                    "default": slots["remote_tools"],
                                }
                            if "local_tools" in slots:
                                opts["allow_local_tools"] = {
                                    "type": "bool",
                                    "label": trans("agent.option.tools.local"),
                                    "description": trans("agent.option.tools.local.desc"),
                                    "default": slots["local_tools"],
                                }
                        tab["options"] = opts
                        options[sub_agent_id] = tab
            except Exception as e:
                self.window.core.debug.log(f"Failed to build options for custom agent '{agent_id}': {e}")
                continue

        # debug tab - trace_id, etc.
        """
        options["debug"] = {
            "label": trans("agent.tab.debug"),
            "options": {
                "trace_id": {
                    "type": "text",
                    "label": trans("agent.option.debug.trace_id"),
                    "description": trans("agent.option.debug.trace_id.desc"),
                    "default": "",
                }
            }
        }
        """
        return options

    def new_agent(self, name: str):
        """
        Create new custom agent

        :param name: agent name
        """
        if not self.loaded:
            self.load()
        new_id = str(uuid4())
        new_agent = AgentItem()
        new_agent.id = new_id
        new_agent.name = name
        self.agents[new_id] = new_agent
        self.save()
        return new_id

    def duplicate_agent(self, agent_id: str, new_name: str):
        """
        Duplicate custom agent

        :param agent_id: agent ID
        :param new_name: new agent name
        """
        if not self.loaded:
            self.load()
        agent = self.agents.get(agent_id)
        if agent:
            new_agent = copy.deepcopy(agent)
            new_agent.id = str(uuid4())
            new_agent.name = new_name
            self.agents[new_agent.id] = new_agent
            self.save()

    def delete_agent(self, agent_id: str):
        """
        Delete custom agent

        :param agent_id: agent ID
        """
        if not self.loaded:
            self.load()
        if agent_id in self.agents:
            del self.agents[agent_id]
            self.save()

    def update_agent(self, agent_id: str, layout: dict, schema: list):
        """
        Update layout and schema of a specific custom agent

        :param agent_id: agent ID
        :param layout: dictionary with new layout
        :param schema: list with new schema
        """
        if not self.loaded:
            self.load()
        agent = self.agents.get(agent_id)
        if agent:
            if layout is None:
                layout = {}
            if schema is None:
                schema = []
            agent.layout = layout
            agent.schema = schema
            self.save()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.29 18:00:00                  #
# ================================================== #

import os
import wave
import base64

from .base import BaseProvider


class GoogleGenAITextToSpeech(BaseProvider):
    def __init__(self, *args, **kwargs):
        """
        Google GenAI Text-to-Speech provider (Gemini TTS via API).

        :param args: args
        :param kwargs: kwargs
        """
        super(GoogleGenAITextToSpeech, self).__init__(*args, **kwargs)
        self.plugin = kwargs.get("plugin")
        self.id = "google_genai_tts"
        self.name = "Google GenAI TTS"

        # Supported preview TTS models (fallback to flash if invalid)
        self.allowed_models = [
            "gemini-2.5-flash-preview-tts",
            "gemini-2.5-pro-preview-tts",
        ]

        # Prebuilt voice names exposed by Gemini TTS
        # Keep list in sync with official docs; fallback to "Puck" if invalid.
        self.allowed_voices = [
            "Zephyr", "Puck", "Charon", "Kore", "Fenrir", "Leda", "Orus",
            "Aoede", "Callirrhoe", "Autonoe", "Enceladus", "Iapetus",
            "Umbriel", "Algieba", "Despina", "Erinome", "Algenib",
            "Rasalgethi", "Laomedeia", "Achernar", "Alnilam", "Schedar",
            "Gacrux", "Pulcherrima", "Achird", "Zubenelgenubi",
            "Vindemiatrix", "Sadachbia", "Sadaltager", "Sulafat",
        ]

    def init_options(self):
        """Initialize options"""
        # Keep option names consistent with the app style; simple text fields are enough.
        self.plugin.add_option(
            "google_genai_tts_model",
            type="text",
            value="gemini-2.5-flash-preview-tts",
            label="Model",
            tab="google_genai_tts",
            description="Specify Gemini TTS model, e.g.: gemini-2.5-flash-preview-tts or gemini-2.5-pro-preview-tts",
        )
        self.plugin.add_option(
            "google_genai_tts_voice",
            type="text",
            value="Kore",
            label="Voice",
            tab="google_genai_tts",
            description="Specify voice, e.g.: Puck, Kore, Charon, Leda, Zephyr... (case-sensitive)",
            urls={
                "Voices": "https://ai.google.dev/gemini-api/docs/speech-generation"
            },
        )

    def speech(self, text: str) -> str:
        """
        Text to speech synthesis using Google GenAI (Gemini TTS).

        :param text: text to synthesize
        :return: path to generated audio file
        """
        # Get pre-configured GenAI client
        client = self.plugin.window.core.api.google.get_client()

        # Resolve path where audio should be written
        output_file = self.plugin.output_file
        path = os.path.join(self.plugin.window.core.config.path, output_file)

        # Validate/select model
        model = self.plugin.get_option_value("google_genai_tts_model") or "gemini-2.5-flash-preview-tts"
        model = self._normalize_model_name(model)
        if model not in self.allowed_models:
            model = "gemini-2.5-flash-preview-tts"

        # Validate/select voice
        voice = self.plugin.get_option_value("google_genai_tts_voice") or "Kore"
        # if voice not in self.allowed_voices:
            # voice = "Kore"

        # Build generation config for audio modality + voice
        # Using explicit types for clarity and forward-compatibility
        try:
            from google.genai import types
        except Exception as ex:
            # Fail fast if SDK is missing or incompatible
            raise RuntimeError("google.genai SDK is not available. Please install/update Google GenAI SDK.") from ex

        gen_config = types.GenerateContentConfig(
            response_modalities=["AUDIO"],
            speech_config=types.SpeechConfig(
                voice_config=types.VoiceConfig(
                    prebuilt_voice_config=types.PrebuiltVoiceConfig(
                        voice_name=voice
                    )
                )
            ),
            temperature=0.8,  # balanced default; keep configurable later if needed
        )

        # Perform TTS request
        response = client.models.generate_content(
            model=model,
            contents=text,
            config=gen_config,
        )

        # Extract PCM bytes from the first candidate/part
        pcm = self._extract_pcm_bytes(response)

        # Persist as standard WAV (PCM 16-bit, mono, 24 kHz)
        self._save_wav(path, pcm, channels=1, rate=24000, sample_width=2)

        return str(path)

    def _extract_pcm_bytes(self, response) -> bytes:
        """
        Extract PCM bytes from generate_content response.

        :param response: Google GenAI response object
        :return: raw PCM byte data
        """
        # Defensive extraction to support minor SDK variations
        data = None
        try:
            cand = response.candidates[0]
            part = cand.content.parts[0]
            if getattr(part, "inline_data", None) and getattr(part.inline_data, "data", None):
                data = part.inline_data.data
        except Exception:
            pass

        if data is None:
            # Some SDK builds may return base64 str; try resolving alternative layout
            try:
                parts = getattr(response, "candidates", [])[0].content.parts
                for p in parts:
                    if getattr(p, "inline_data", None) and getattr(p.inline_data, "data", None):
                        data = p.inline_data.data
                        break
            except Exception:
                pass

        if data is None:
            raise RuntimeError("No audio data returned by Gemini TTS response.")

        # Normalize to raw bytes
        if isinstance(data, (bytes, bytearray)):
            return bytes(data)
        if isinstance(data, str):
            # Fallback: treat as base64-encoded PCM
            return base64.b64decode(data)

        # Last resort: try bytes() cast
        try:
            return bytes(data)
        except Exception as ex:
            raise RuntimeError("Unsupported audio payload type returned by Gemini TTS.") from ex

    def _save_wav(
            self,
            filename: str,
            pcm_bytes: bytes,
            channels: int = 1,
            rate: int = 24000,
            sample_width: int = 2
    ):
        """
        Save raw PCM bytes to a WAV file.

        :param filename: output WAV file path
        :param pcm_bytes: raw PCM byte data
        :param channels: number of audio channels (1=mono, 2=stereo)
        :param rate: sample rate in Hz (e.g., 24000)
        :param sample_width: sample width in bytes (e.g., 2 for 16-bit)
        """
        # Ensure parent directory exists
        os.makedirs(os.path.dirname(filename), exist_ok=True)

        # Write PCM payload as WAV
        with wave.open(filename, "wb") as wf:
            wf.setnchannels(channels)
            wf.setsampwidth(sample_width)  # bytes per sample (2 -> 16-bit)
            wf.setframerate(rate)
            wf.writeframes(pcm_bytes)

    def _normalize_model_name(self, model: str) -> str:
        """
        Normalize model id (strip optional 'models/' prefix).

        :param model: model id
        """
        try:
            return model.split("/")[-1]
        except Exception:
            return model

    def is_configured(self) -> bool:
        """
        Check if provider is configured

        :return: True if configured, False otherwise
        """
        api_key = self.plugin.window.core.config.get("api_key_google")
        return api_key is not None and api_key != ""

    def get_config_message(self) -> str:
        """
        Return message to display when provider is not configured

        :return: message
        """
        return "Google GenAI API key is not set yet. Please configure it in settings."


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.27 21:00:00                  #
# ================================================== #

from typing import Optional, List, Union

from PySide6.QtCore import QModelIndex, QTimer
from PySide6.QtGui import QStandardItem

from pygpt_net.item.ctx import CtxItem, CtxMeta
from pygpt_net.core.types import MODE_ASSISTANT
from pygpt_net.core.events import (
    Event,
    AppEvent,
    RenderEvent,
    BaseEvent
)
from pygpt_net.utils import trans

from .common import Common
from .summarizer import Summarizer
from .extra import Extra

class Ctx:
    def __init__(self, window=None):
        """
        Context controller

        :param window: Window instance
        """
        self.window = window
        self.common = Common(window)
        self.summarizer = Summarizer(window)
        self.extra = Extra(window)

        # current edit IDs
        self.edit_meta_id = None
        self.edit_item_id = None

        # current group ID
        self.group_id = None
        self.selected = []
        self._infinite_scroll_refresh = False

    def handle(self, event: BaseEvent):
        """
        Handle events

        :param event: BaseEvent: Event to handle
        """
        name = event.name

        # on input begin
        if name == Event.INPUT_BEGIN:
            force = event.data.get("force", False)
            stop = event.data.get("stop", False)
            if not force and not stop:
                if self.extra.is_editing():
                    event.data["stop"] = True  # stop flow
                    self.extra.edit_submit()
                    return

        # on input before
        elif name == Event.INPUT_BEFORE:
            mode = event.data.get("mode")
            text = event.data.get("value", "")
            multimodal_ctx = event.data.get("multimodal_ctx", None)
            # check if image captured from camera and attachment exists
            camera_captured = (self.window.controller.ui.vision.has_vision()
                               and self.window.controller.attachment.has(mode))

            # allow empty text input only if multimodal data, otherwise abort
            is_audio = multimodal_ctx is not None and multimodal_ctx.is_audio_input
            if len(text.strip()) == 0 and (not camera_captured and not is_audio):
                event.data["stop"] = True  # stop flow
                event.data["silent"] = True  # silent stop (no errors)
                return

        # on input accept
        elif name == Event.INPUT_ACCEPT:
            mode = event.data.get("mode")
            # prepare ctx, create new ctx meta if there is no ctx, or no ctx selected
            if self.window.core.ctx.count_meta() == 0 or self.window.core.ctx.get_current() is None:
                self.window.core.ctx.new()
                self.update()
                self.window.controller.chat.log("New context created...")  # log
            else:
                # check if current ctx is allowed for this mode - if not, then auto-create new ctx
                self.handle_allowed(mode)

            # update mode in ctx
            self.update_mode_in_current()

    def setup(self):
        """Setup ctx"""
        try:
            self.common.restore_display_filter()  # load filters first
        except Exception:
            pass

        # load ctx list
        core = self.window.core
        core.ctx.load_meta()

        # if no context yet then create one
        if core.ctx.count_meta() == 0:
            self.new()
        else:
            id = core.config.get('ctx')
            if id is not None:
                # Keep previously selected id; if it's not in the current page
                # we will inject a placeholder into the list (see update_list).
                core.ctx.set_current(id)
            else:
                core.ctx.set_current(core.ctx.get_first())

        # restore search string if exists
        if self.window.core.config.has("ctx.search.string"):
            string = self.window.core.config.get("ctx.search.string")
            if string:
                self.window.ui.nodes['ctx.search'].setText(string)
                self.search_string_change(string)
                # check if current selected ctx is still valid
                if self.window.core.ctx.get_current() is not None:
                    if not self.window.core.ctx.has(self.window.core.ctx.get_current()):
                        self.search_string_clear()

        self.window.ui.nodes['ctx.list'].collapseAll()  # collapse all items at start
        self.restore_expanded_groups()  # restore expanded groups
        self.select_by_current(focus=True)  # scroll to current ctx

        # focus input after loading
        QTimer.singleShot(2000, self.window.controller.chat.common.focus_input)

    def update_mode_in_current(self):
        """Update current ctx mode"""
        mode = self.window.core.config.get('mode')
        model = self.window.core.config.get('model')
        id = self.window.core.ctx.get_current()
        if id is not None:
            meta = self.window.core.ctx.get_meta_by_id(id)
            if meta:
                self.window.core.ctx.mode = mode
                self.window.core.ctx.model = model
                self.window.core.ctx.last_mode = mode
                self.window.core.ctx.last_model = model
                meta.mode = mode
                meta.model = model
                meta.last_mode = mode
                meta.last_model = model
                self.window.core.ctx.save(id)

    def update(
            self,
            reload: bool = True,
            all: bool = True,
            select: bool = True,
            no_scroll: bool = False,
    ):
        """
        Update ctx list

        :param reload: reload ctx list items
        :param all: update all
        :param select: select current ctx
        :param no_scroll: do not scroll to selected item
        """
        if reload:
            self.update_list(True)

        if select:
            nodes = self.window.ui.nodes
            if no_scroll:
                nodes['ctx.list'].store_scroll_position()
            self.select_by_current()
            if no_scroll:
                nodes['ctx.list'].restore_scroll_position()

        if all:
            self.window.controller.ui.update()

        id = self.window.core.ctx.get_current()
        if id is not None:
            self.window.core.config.set('ctx', id)
            self.window.core.config.set('assistant_thread', self.window.core.ctx.get_thread())
            self.window.core.config.save()

        self.window.controller.calendar.update(all=False)
        self.window.controller.chat.attachment.update()

    def select(
            self,
            id: int,
            force: bool = False
    ):
        """
        Select ctx by id

        :param id: context meta id
        :param force: force select
        """
        prev_id = self.window.core.ctx.get_current()
        self.window.core.ctx.set_current(id)
        meta = self.window.core.ctx.get_meta_by_id(id)
        if prev_id != id or force:
            self.load(id)
            self.window.dispatch(AppEvent(AppEvent.CTX_SELECTED))  # app event
        else:
            if meta is not None:
                self.set_group(meta.group_id)

        self.common.focus_chat(meta)
        self.window.controller.chat.attachment.update()
        self.set_selected(id)
        self.clean_memory()  # clean memory

    def select_on_list_only(self, id: int):
        """
        Select ctx by id only on list

        :param id: context meta id
        """
        ctx = self.window.core.ctx
        ctx.select(id, restore_model=True)
        ctx.set_current(id)
        self.update_list(True)
        self.select_by_current()
        self.reload_config(all=False)
        self.update(reload=False, all=True, select=False)
        self.set_selected(id)

    def select_by_idx(self, idx: int):
        """
        Select ctx by index

        :param idx: context index
        """
        self.select_by_id(self.window.core.ctx.get_id_by_idx(idx))

    def select_by_id(self, id: int):
        """
        Select ctx by index

        :param id: context index
        """
        if self.context_change_locked():
            return

        self.select(id)
        event = Event(Event.CTX_SELECT, {
            'value': id,
        })
        self.window.dispatch(event)

    def select_by_current(self, focus: bool = False):
        """
        Select ctx by current

        :param focus: focus chat
        """
        id = self.window.core.ctx.get_current()
        meta = self.window.core.ctx.get_meta()
        if id in meta:
            self.select_index_by_id(id)
        if focus:
            self.update()
            index = self.get_child_index_by_id(id)
            if index.isValid():
                self.window.ui.nodes['ctx.list'].force_scroll_to_current()

    def unselect(self):
        """Unselect ctx"""
        self.set_group(None)
        self.window.ui.nodes['ctx.list'].clearSelection()
        self.clear_selected()

    def set_group(self, group_id: Optional[int] = None):
        """
        Set current selected group

        :param group_id: group ID
        """
        self.group_id = group_id

    def search_focus_in(self):
        """Search focus handler"""
        pass
        # self.select_by_current()

    def new_ungrouped(self):
        """Create new ungrouped ctx"""
        self.group_id = None
        self.new()

    def clean_memory(self):
        """Clean memory"""
        pass

    def new_in_group(
            self,
            group_id: Optional[int] = None,
            force: bool = False
    ):
        """
        Create new ctx in group

        :param group_id: group ID
        :param force: force context creation
        """
        QTimer.singleShot(10, lambda: self.new(force, group_id))

    def new(
            self,
            force: bool = False,
            group_id: Optional[Union[int, list]] = None
    ):
        """
        Create new ctx

        :param force: force context creation
        :param group_id: group ID
        """
        if not force and self.context_change_locked():
            return

        # if multiple group IDs provided, pick the first valid
        if isinstance(group_id, list):
            valid_group_id = None
            for gid in group_id:
                if self.window.core.ctx.has_group(gid):
                    valid_group_id = gid
                    break
            group_id = valid_group_id

        if group_id is None:
            if self.group_id is not None and self.group_id > 0:
                group_id = self.group_id
        else:
            self.group_id = group_id

        if group_id is not None and not self.window.core.ctx.has_group(group_id):
            group_id = None
            self.group_id = None

        meta = self.window.core.ctx.new(group_id)
        self.window.core.config.set('assistant_thread', None)  # reset assistant thread id
        self.update()

        self.fresh_output(meta)  # render reset

        if not force:
            self.window.controller.chat.common.unlock_input()
            self.window.controller.chat.common.focus_input()

        mode = self.window.core.ctx.get_mode()
        assistant_id = None
        if mode == MODE_ASSISTANT:
            assistant_id = self.window.core.config.get('assistant')
            self.window.controller.assistant.files.update()  # always update assistant files

        self.common.update_label(mode, assistant_id)
        self.common.focus_chat(meta)

        if meta is not None:
            self.window.controller.ui.tabs.update_title_current(meta.name)

        self.window.dispatch(AppEvent(AppEvent.CTX_CREATED))
        self.select(meta.id)
        QTimer.singleShot(1000, self.window.controller.chat.common.focus_input)
        return meta

    def add(self, ctx: CtxItem):
        """
        Add ctx item (CtxItem object)

        :param ctx: CtxItem
        """
        self.window.core.ctx.add(ctx)
        self.update()

    def prev(self):
        """Select previous ctx"""
        id = self.window.core.ctx.get_prev()
        if id is not None:
            self.select(id)

    def next(self):
        """Select next ctx"""
        id = self.window.core.ctx.get_next()
        if id is not None:
            self.select(id)

    def last(self):
        """Select last (newest) ctx"""
        id = self.window.core.ctx.get_last_meta()
        if id is not None:
            self.select(id)

    def update_list(self, reload: bool = False, restore_scroll: bool = False):
        """
        Reload current ctx list

        :param reload: reload ctx list items
        :param restore_scroll: restore scroll position
        """
        view = self.window.ui.nodes['ctx.list']

        # Read raw scrollbar state BEFORE model rebuild
        sb = None
        prev_val = None
        if restore_scroll:
            try:
                sb = view.verticalScrollBar()
                prev_val = sb.value()
            except Exception:
                sb = None
                prev_val = None

            # In infinite-scroll mode do NOT call store/restore helpers to avoid collisions.
            # Instead, schedule a pending scroll that will be applied while updates are disabled.
            if self._infinite_scroll_refresh:
                if prev_val is not None:
                    try:
                        view.set_pending_v_scroll(prev_val)
                    except Exception:
                        pass
            else:
                # Normal mode: keep existing helpers
                try:
                    view.store_scroll_position()
                except Exception:
                    pass

        # Prepare data (with placeholder injection if current is outside of the page)
        data = self.window.core.ctx.get_meta(reload)
        data = self._inject_current_if_missing(data)

        # Rebuild list (CtxList.update will apply pending scroll before enabling updates)
        self.window.ui.contexts.ctx_list.update(
            'ctx.list',
            data,
            expand=False,
        )

        # Post-fix: in normal mode let helpers restore; in infinite mode we already applied pending
        if restore_scroll:
            if self._infinite_scroll_refresh:
                # Optional safety re-apply to the same value after layout settles
                if sb is not None and prev_val is not None:
                    QTimer.singleShot(0, lambda: sb.setValue(prev_val))
            else:
                try:
                    view.restore_scroll_position()
                    QTimer.singleShot(0, view.restore_scroll_position)
                except Exception:
                    pass

    def refresh(self, restore_model: bool = True):
        """
        Refresh context

        :param restore_model: restore model
        """
        self.load(
            self.window.core.ctx.get_current(),
            restore_model,
        )

    def refresh_output(self):
        """Refresh output"""
        data = {
            "meta": self.window.core.ctx.get_current_meta(),
            "items": self.window.core.ctx.get_items(),
            "clear": True,
        }
        event = RenderEvent(RenderEvent.CTX_APPEND, data)
        self.window.dispatch(event)

    def load(
            self,
            id: int,
            restore_model: bool = True,
            select_idx: Optional[int] = None,
            new_tab: Optional[bool] = False,
            no_fresh: bool = False
    ):
        """
        Load ctx data

        :param id: context ID
        :param restore_model: restore model if defined in ctx
        :param select_idx: select index on list after loading
        :param new_tab: open in new tab
        :param no_fresh: do not fresh output
        """
        if new_tab:
            col_idx = self.window.controller.ui.tabs.column_idx
            self.window.controller.ui.tabs.create_new_on_tab = False
            self.window.controller.ui.tabs.new_tab(col_idx)

        self.window.core.ctx.clear_thread()
        self.window.core.ctx.select(id, restore_model=restore_model)
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is not None:
            self.set_group(meta.group_id)

        if meta is not None and not no_fresh:
            self.fresh_output(meta)

        self.reload_config()
        self.update(reload=False, all=True)

        if meta is not None:
            self.window.controller.ui.tabs.on_load_ctx(meta)

        if select_idx is not None:
            self.select(id)
            self.window.ui.nodes['ctx.list'].select_by_idx(select_idx)

    def after_load(
            self,
            id: int,
    ):
        """
        Load ctx data

        :param id: context ID
        """
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is not None:
            self.set_group(meta.group_id)

        if meta is not None:
            data = {
                "meta": meta,
            }
            event = RenderEvent(RenderEvent.ON_LOAD, data)
            self.window.dispatch(event)

        self.reload_config()
        self.update(reload=False, all=True)

        if meta is not None:
            self.window.controller.ui.tabs.on_load_ctx(meta)

    def reload_config(self, all: bool = True):
        """
        Reload config

        :param all: reload all
        """
        ctx = self.window.core.ctx
        ctrl = self.window.controller

        thread = ctx.get_thread()
        mode = ctx.get_mode()
        model = ctx.get_model()
        assistant_id = ctx.get_assistant()
        preset = ctx.get_preset()

        self.window.core.config.set('assistant_thread', thread)

        if all:
            self.refresh_output()

        if mode is not None:
            ctrl.mode.set(mode)

            if preset is not None:
                ctrl.presets.set(mode, preset)
                ctrl.presets.refresh()

            if mode == MODE_ASSISTANT:
                if assistant_id is not None:
                    ctrl.assistant.select_by_id(assistant_id)
                else:
                    assistant_id = self.window.core.config.get('assistant')
                ctrl.assistant.files.update()

            if model is not None and self.window.core.models.has_model(mode, model):
                ctrl.model.set(mode, model)

        self.common.update_label(mode, assistant_id)

    def update_ctx(self):
        """Update current ctx mode if allowed"""
        mode = self.window.core.config.get('mode')
        id = None
        if self.window.core.ctx.is_allowed_for_mode(mode, False):
            self.window.core.ctx.update()
            if mode == MODE_ASSISTANT:
                if self.window.core.ctx.get_assistant() is not None:
                    id = self.window.core.ctx.get_assistant()
                else:
                    id = self.window.core.config.get('assistant')

        self.common.update_label(mode, id)

    def delete(
            self,
            id: Union[int, list],
            force: bool = False
    ):
        """
        Delete ctx by idx

        :param id: context meta idx on list or list of idxs
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.delete',
                id=id,
                msg=trans('ctx.delete.confirm'),
            )
            return
        updated = False
        updated_current = False
        ids = id if isinstance(id, list) else [id]
        for id in ids:
            try:
                self.delete_meta_from_idx(id)
            except Exception as e:
                self.window.core.debug.log(e)
                print("Error deleting ctx data from indexes", e)

            if self.window.core.ctx.get_current() == id:
                items = self.window.core.ctx.all()  # TODO: get by meta id(s)
                self.window.core.history.remove_items(items)
                updated_current = True
            self.window.core.attachments.context.delete_by_meta_id(id)
            self.window.core.ctx.remove(id)
            self.remove_selected(id)

            if self.window.core.ctx.get_current() == id:
                self.window.core.ctx.clear_current()
                event = RenderEvent(RenderEvent.CLEAR_OUTPUT)
                self.window.dispatch(event)
            updated = True

        if updated:
            self.update_and_restore()
            if updated_current:
                self.window.controller.ui.tabs.update_title_current("...")

    def delete_meta_from_idx(self, id: int):
        """
        Delete meta from indexes

        :param id: ctx meta id
        """
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is None:
            return

        if meta.indexed is not None and meta.indexed > 0:
            for store_id in list(meta.indexes):
                for idx in list(meta.indexes[store_id]):
                    self.window.core.ctx.idx.remove_meta_from_indexed(store_id, id, idx)

    def delete_item(
            self,
            id: int,
            force: bool = False
    ):
        """
        Delete ctx item by id

        :param id: ctx item id
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.delete_item',
                id=id,
                msg=trans('ctx.delete.item.confirm'),
            )
            return
        item = self.window.core.ctx.get_item_by_id(id)
        event = RenderEvent(RenderEvent.ITEM_DELETE_ID, {
            "ctx": item,
        })
        self.window.core.ctx.remove_item(id)
        self.window.dispatch(event)

    def delete_history(self, force: bool = False):
        """
        Delete all ctx / truncate

        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.delete_all',
                id='',
                msg=trans('ctx.delete.all.confirm'),
            )
            return

        try:
            self.window.core.idx.ctx.truncate()
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error truncating ctx index db", e)

        self.group_id = None
        self.unselect()
        self.window.core.ctx.truncate()
        self.window.core.history.truncate()
        self.window.core.attachments.context.truncate()
        self.clear_selected()
        self.update()
        self.new()

    def delete_history_groups(self, force: bool = False):
        """
        Delete all ctx / truncate

        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.delete_all_groups',
                id='',
                msg=trans('ctx.delete.all.confirm'),
            )
            return

        try:
            self.window.core.idx.ctx.truncate()
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error truncating ctx index db", e)

        self.group_id = None
        self.unselect()
        self.window.core.ctx.truncate()
        self.window.core.history.truncate()
        self.window.core.ctx.truncate_groups()
        self.window.core.attachments.context.truncate()
        self.clear_selected()
        self.update()
        self.new()

    def new_if_empty(self):
        """Create new context if empty"""
        self.group_id = None
        if self.window.core.ctx.count_meta() == 0:
            self.new()

    def rename(self, id: Union[int, list]):
        """
        Ctx name rename by id (show dialog)

        :param id: context id or list of ids
        """
        first_id = id
        if isinstance(id, list):
            if len(id) == 0:
                return
            first_id = id[0]
        meta = self.window.core.ctx.get_meta_by_id(first_id)
        self.window.ui.dialog['rename'].id = 'ctx'
        self.window.ui.dialog['rename'].input.setText(meta.name)
        self.window.ui.dialog['rename'].current = id
        self.window.ui.dialog['rename'].show()

    def set_important(
            self,
            id: Union[int, list],
            value: bool = True
    ):
        """
        Set as important

        :param id: context idx or list of idxs
        :param value: important value
        """
        updated = False
        ids = id if isinstance(id, list) else [id]
        for id in ids:
            meta = self.window.core.ctx.get_meta_by_id(id)
            if meta is not None:
                meta.important = value
                self.window.core.ctx.save(id)
                updated = True
        if updated:
            self.update_and_restore()

    def is_important(self, idx: int) -> bool:
        """
        Check if ctx is important

        :param idx: context idx
        :return: True if important
        """
        id = self.window.core.ctx.get_id_by_idx(idx)
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is not None:
            return meta.important
        return False

    def set_label(
            self,
            id: Union[int, list],
            label_id: int
    ):
        """
        Set color label for ctx by idx

        :param id: context idx or list of idxs
        :param label_id: label id
        """
        updated = False
        ids = id if isinstance(id, list) else [id]
        for id in ids:
            meta = self.window.core.ctx.get_meta_by_id(id)
            if meta is not None:
                meta.label = label_id
                self.window.core.ctx.save(id)
                updated = True

        if updated:
            QTimer.singleShot(
                10,
                lambda: self.update_and_restore()
            )

    def update_and_restore(self):
        """Update ctx and restore scroll position"""
        self.window.ui.nodes['ctx.list'].store_scroll_position()
        self.update()
        self.window.ui.nodes['ctx.list'].restore_scroll_position()

    def update_name(
            self,
            id: Union[int, list],
            name: str,
            close: bool = True,
            refresh: bool = True
    ):
        """
        Update ctx name

        :param id: context id or list of ids
        :param name: context name
        :param close: close rename dialog
        :param refresh: refresh ctx list
        """
        ctx_id = id
        if not isinstance(id, list):
            ctx_id = [id]

        for id in ctx_id:
            if id not in self.window.core.ctx.get_meta():
                continue
            self.window.core.ctx.meta[id].name = name
            self.window.core.ctx.set_initialized()
            self.window.core.ctx.save(id)

        if close:
            self.window.ui.dialog['rename'].close()

        if refresh:
            self.update_and_restore()
        else:
            self.update(reload=True, all=False, no_scroll=True)

        for id in ctx_id:
            if id not in self.window.core.ctx.get_meta():
                continue
            meta = self.window.core.ctx.get_meta_by_id(id)
            if meta is not None:
                if id == self.window.core.ctx.get_current():
                    self.window.controller.ui.tabs.update_title_current(meta.name)

    def update_name_current(self, name: str):
        """
        Update current ctx name

        :param name: context name
        """
        self.update_name(self.window.core.ctx.get_current(), name)

    def handle_allowed(self, mode: str) -> bool:
        """
        Check if append to current ctx is allowed for this mode, if not then switch to new context

        :param mode: mode name
        :return: True if allowed
        """
        if not self.window.core.ctx.is_allowed_for_mode(mode):
            self.new(True)  # force new context
            return False
        return True

    def selection_change(self):
        """Select ctx on list change"""
        selected_idx = self.window.ui.nodes['ctx.list'].currentIndex()
        if selected_idx.isValid():
            _ = self.window.core.ctx.get_id_by_idx(selected_idx.row())
        self.window.ui.nodes['ctx.list'].lockSelection()

    def search_string_change(self, text: str):
        """
        Search string changed handler

        :param text: search string
        """
        self.reset_loaded_total()  # reset paging on new search
        self.window.core.ctx.clear_tmp_meta()
        self.window.core.ctx.set_search_string(text)
        self.window.core.config.set('ctx.search.string', text)
        self.update(reload=True, all=False)

    def search_string_clear(self):
        """Search string clear"""
        self.window.ui.nodes['ctx.search'].clear()
        self.search_string_change("")

    def append_search_string(self, text: str):
        """
        Append search string to input and make search

        :param text: search string
        """
        self.window.ui.nodes['ctx.search'].setText(text)
        self.search_string_change(text)

    def label_filters_changed(self, labels: List[int]):
        """
        Filters labels change

        :param labels: list of labels
        """
        self.reset_loaded_total()  # reset paging on label filter change
        self.window.core.ctx.clear_tmp_meta()
        self.window.core.ctx.filters_labels = labels
        self.window.core.config.set('ctx.records.filter.labels', labels)
        self.update(reload=True, all=False)

    def prepare_name(self, ctx: CtxItem, force: bool = False):
        """
        Handle context name (summarize first input and output)

        :param ctx: CtxItem
        :param force: force update
        """
        if not self.window.core.ctx.is_initialized() or force:
            self.summarizer.summarize(
                self.window.core.ctx.get_current(),
                ctx,
            )

    def context_change_locked(self) -> bool:
        """
        Check if ctx change is locked

        :return: True if locked
        """
        return self.window.controller.chat.input.generating

    def select_index_by_id(self, id: int):
        """
        Select item by ID on context list

        :param id: ctx meta ID
        """
        index = self.get_child_index_by_id(id)
        nodes = self.window.ui.nodes
        nodes['ctx.list'].unlocked = True
        if not index.isValid():
            nodes['ctx.list'].unlocked = False
            return
        nodes['ctx.list'].setCurrentIndex(index)
        nodes['ctx.list'].force_scroll_to_current()
        nodes['ctx.list'].unlocked = False

    def find_index_by_id(
            self,
            item: QStandardItem,
            id: int
    ) -> QModelIndex:
        """
        Return index of item with given ID, searching recursively through the model.

        :param item: QStandardItem
        :param id: int
        :return: QModelIndex
        """
        finder = self.find_index_by_id
        if hasattr(item, 'id') and item.id == id:
            return item.index()
        for row in range(item.rowCount()):
            found_index = finder(item.child(row), id)
            if found_index.isValid():
                return found_index
        return QModelIndex()

    def find_parent_index_by_id(
            self,
            item: QStandardItem,
            id: int
    ) -> QModelIndex:
        """
        Return index of item with given ID, searching recursively through the model.

        :param item: QStandardItem
        :param id: int
        :return: QModelIndex
        """
        finder = self.find_parent_index_by_id
        if hasattr(item, 'id') and hasattr(item, 'isFolder') and item.isFolder and item.id == id:
            return item.index()
        for row in range(item.rowCount()):
            found_index = finder(item.child(row), id)
            if found_index.isValid():
                return found_index
        return QModelIndex()

    def get_parent_index_by_id(self, id: int) -> QModelIndex:
        """
        Return QModelIndex of parent item based on its ID.

        :param id: int
        :return: QModelIndex
        """
        model = self.window.ui.models['ctx.list']
        root = model.invisibleRootItem()
        return self.find_parent_index_by_id(root, id)

    def get_children_index_by_id(
            self,
            parent_id: int,
            child_id: int
    ) -> QModelIndex:
        """
        Return QModelIndex of child item based on its ID and parent ID.

        :param parent_id: int
        :param child_id: int
        :return: QModelIndex
        """
        model = self.window.ui.models['ctx.list']
        parent_index = self.get_parent_index_by_id(parent_id)
        if not parent_index.isValid():
            return QModelIndex()

        parent_item = model.itemFromIndex(parent_index)
        return self.find_index_by_id(parent_item, child_id)

    def find_child_index_by_id(
            self,
            root_item: QStandardItem,
            child_id: int
    ) -> QModelIndex:
        """
        Find and return QModelIndex of child based on its ID, recursively searching through the model.

        :param root_item: QStandardItem
        :param child_id: int
        :return: QModelIndex
        """
        finder = self.find_child_index_by_id
        for row in range(root_item.rowCount()):
            item = root_item.child(row)
            if hasattr(item, 'id') and hasattr(item, 'isFolder') and not item.isFolder and item.id == child_id:
                return item.index()
            child_index = finder(item, child_id)
            if child_index.isValid():
                return child_index
        return QModelIndex()

    def get_child_index_by_id(self, child_id: int) -> QModelIndex:
        """
        Return QModelIndex of child item based on its ID.

        :param child_id: int
        :return: QModelIndex
        """
        model = self.window.ui.models['ctx.list']
        root_item = model.invisibleRootItem()
        return self.find_child_index_by_id(root_item, child_id)

    def store_expanded_groups(self):
        """
        Store expanded groups in ctx list
        """
        items = self.window.ui.nodes['ctx.list'].expanded_items
        self.window.core.config.set('ctx.list.expanded', list(items))
        self.window.core.config.save()

    def restore_expanded_groups(self):
        """
        Restore expanded groups in ctx list
        """
        expanded = self.window.core.config.get('ctx.list.expanded')
        if expanded is not None:
            nodes = self.window.ui.nodes
            for group_id in expanded:
                nodes['ctx.list'].expand_group(group_id)

    def save_all(self):
        """Save visible ctx list items"""
        self.store_expanded_groups()

    def move_to_group(
            self,
            meta_id: Union[int, list],
            group_id: int,
            update: bool = True
    ):
        """
        Move ctx to group

        :param meta_id: int or list of int
        :param group_id: int
        :param update: update ctx list
        """
        updated = False
        ids = meta_id if isinstance(meta_id, list) else [meta_id]
        for meta_id in ids:
            self.window.core.ctx.update_meta_group_id(meta_id, group_id)
            self.group_id = group_id
            updated = True

        if updated and update:
            QTimer.singleShot(
                10,
                lambda: self.update_and_restore()
            )

    def remove_from_group(self, meta_id: Union[int, list]):
        """
        Remove ctx from group

        :param meta_id: int or list of int
        """
        updated = False
        ids = meta_id if isinstance(meta_id, list) else [meta_id]
        for meta_id in ids:
            self.window.core.ctx.update_meta_group_id(meta_id, None)
            updated = True
        if updated:
            self.group_id = None
            QTimer.singleShot(
                10,
                lambda: self.update_and_restore()
            )

    def new_group(
            self,
            meta_id: Optional[int] = None
    ):
        """
        Open new group dialog

        :param meta_id: int
        """
        self.window.ui.dialog['create'].id = 'ctx.group'
        self.window.ui.dialog['create'].input.setText("")
        self.window.ui.dialog['create'].current = meta_id
        self.window.ui.dialog['create'].show()
        self.window.ui.dialog['create'].input.setFocus()

    def create_group(
            self,
            name: Optional[str] = None,
            meta_id: Optional[Union[int, list]] = None
    ):
        """
        Make directory

        :param name: name of directory
        :param meta_id: int or list of int
        """
        if name is None:
            self.window.update_status(
                "[ERROR] Name is empty."
            )
            return
        group = self.window.core.ctx.make_group(name)
        id = self.window.core.ctx.insert_group(group)
        if id is not None:
            ids = meta_id if isinstance(meta_id, list) else [meta_id] if meta_id is not None else []
            for meta_id in ids:
                if meta_id is not None:
                    self.move_to_group(meta_id, id, update=False)

        self.update()
        self.window.update_status(
            "Group '{}' created.".format(name)
        )
        self.window.ui.dialog['create'].close()
        # self.select_group(id)
        self.group_id = id

    def rename_group(
            self,
            id: Union[int, list],
            force: bool = False
    ):
        """
        Rename group

        :param id: group ID or list of IDs
        :param force: force rename
        """
        ids = id if isinstance(id, list) else [id]
        if not force:
            is_group = False
            name = ""
            for tmp_id in ids:
                group = self.window.core.ctx.get_group_by_id(tmp_id)
                if group is not None:
                    is_group = True
                    name = group.name
                    break
            if not is_group:
                return
            self.window.ui.dialog['rename'].id = 'ctx.group'
            self.window.ui.dialog['rename'].input.setText(name)
            self.window.ui.dialog['rename'].current = id
            self.window.ui.dialog['rename'].show()

    def update_group_name(
            self,
            id: Union[int, list],
            name: str,
            close: bool = True
    ):
        """
        Update group name

        :param id: group ID or list of IDs
        :param name: group name
        :param close: close rename dialog
        """
        updated = False
        ids = id if isinstance(id, list) else [id]
        for id in ids:
            group = self.window.core.ctx.get_group_by_id(id)
            if group is not None:
                group.name = name
                self.window.core.ctx.update_group(group)
                updated = True
        if updated:
            if close:
                self.window.ui.dialog['rename'].close()
            self.update_and_restore()

    def get_group_name(self, id: int) -> str:
        """
        Get group name by ID

        :param id: group ID
        :return: group name
        """
        group = self.window.core.ctx.get_group_by_id(id)
        if group is not None:
            return group.name
        return ""

    def select_group(self, id: int):
        """
        Select group

        :param id: group ID
        """
        self.group_id = id
        index = self.get_parent_index_by_id(id)
        nodes = self.window.ui.nodes
        nodes['ctx.list'].unlocked = True
        nodes['ctx.list'].setCurrentIndex(index)
        nodes['ctx.list'].force_scroll_to_current()
        nodes['ctx.list'].unlocked = False

    def delete_group(
            self,
            id: Union[int, list],
            force: bool = False
    ):
        """
        Delete group only

        :param id: group ID or list of IDs
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.group.delete',
                id=id,
                msg=trans('confirm.ctx.delete')
            )
            return
        updated = False
        ids = id if isinstance(id, list) else [id]
        for id in ids:
            group = self.window.core.ctx.get_group_by_id(id)
            if group is not None:
                self.window.core.ctx.remove_group(group, all=False)
                if self.group_id == id:
                    self.group_id = None
                updated = True

        if updated:
            self.update_and_restore()

    def delete_group_all(
            self,
            id: Union[int, list],
            force: bool = False
    ):
        """
        Delete group with all items

        :param id: group ID or list of IDs
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.group.delete.all',
                id=id,
                msg=trans('confirm.ctx.delete.all')
            )
            return
        updated = False
        ids = id if isinstance(id, list) else [id]
        for id in ids:
            group = self.window.core.ctx.get_group_by_id(id)
            if group is not None:
                self.window.core.ctx.remove_group(group, all=True)
                if self.group_id == id:
                    self.group_id = None
                updated = True
        if updated:
            self.update_and_restore()

    def prepare_summary(self, ctx: CtxItem) -> bool:
        """
        Prepare context summary

        :param ctx: CtxItem
        :return: True if handled
        """
        if not ctx.meta or not ctx.meta.initialized:  # don't call if reply or internal mode
            if self.window.core.config.get('ctx.auto_summary'):
                self.window.controller.chat.log("Calling for prepare context name...")
                self.prepare_name(ctx)  # async
                return True
        return False

    def store_history(self, ctx: CtxItem, type: str) -> bool:
        """
        Store ctx in history if enabled

        :param ctx: CtxItem
        :param type: input|output
        :return: Tru if stored
        """
        # store to history
        if self.window.core.config.get('store_history'):
            self.window.core.history.append(ctx, type)
            return True
        return False

    def reload(self):
        """Reload ctx"""
        self.edit_meta_id = None
        self.edit_item_id = None
        self.group_id = None
        self.selected = []
        self._infinite_scroll_refresh = False
        self.reset_loaded_total()  # reset paging
        self.window.core.ctx.reset()
        self.setup()

    def reload_after(self):
        """After reload"""
        self.update()
        self.refresh()
        self.new_if_empty()

    def add_selected(self, id: int):
        """
        Add selection ID to selected list

        :param id: context meta ID
        """
        if id not in self.selected:
            self.selected.append(id)

    def remove_selected(self, id: int):
        """
        Remove selection ID from selected list

        :param id: context meta ID
        """
        if id in self.selected:
            self.selected.remove(id)

    def set_selected(self, id: int):
        """
        Set selected ID in selected list

        :param id: context meta ID
        """
        self.selected = [id] if id is not None else []

    def set_selected_by_idx(self, idx: int):
        """
        Set selected ID by index

        :param idx: context meta index
        """
        id = self.window.core.ctx.get_id_by_idx(idx)
        if id is not None:
            self.set_selected(id)

    def clear_selected(self):
        """Clear selected list"""
        self.selected.clear()

    def fresh_current_output(self):
        """Fresh output for current context"""
        meta = self.window.core.ctx.get_meta_by_id(self.window.core.ctx.get_current())
        if meta is not None:
            self.fresh_output(meta)

    def fresh_output(self, meta: CtxMeta):
        """
        Fresh output for new context

        :param meta: CtxItem
        """
        data = {
            "meta": meta,
        }
        event = RenderEvent(RenderEvent.FRESH, data)
        self.window.dispatch(event)
        data = {
            "meta": meta,
        }
        event = RenderEvent(RenderEvent.ON_LOAD, data)
        self.window.dispatch(event)

    def get_package_limit(self) -> int:
        """Return base package size from config."""
        try:
            return int(self.window.core.config.get('ctx.records.limit') or 0)
        except Exception:
            return 0

    def get_loaded_total_limit(self) -> int:
        """Return persisted total loaded size; fallback to base package size."""
        base = self.get_package_limit()
        try:
            val = int(self.window.core.config.get('ctx.records.limit.total') or 0)
        except Exception:
            val = 0
        return val if val > 0 else base

    def reset_loaded_total(self):
        """Reset total loaded to base package; used on filter/search changes."""
        base = self.get_package_limit()
        # If base is 0 (unlimited), keep total as 0 which means "unlimited"
        self.window.core.config.set('ctx.records.limit.total', base if base > 0 else 0)
        self.window.core.config.save()

    def load_more(self, packages: int = 1):
        """
        Increase total loaded limit and append new ungrouped records without rebuilding the list.
        """
        base = self.get_package_limit()
        if base <= 0:
            # Unlimited mode – nothing to do
            return

        current_total = self.get_loaded_total_limit()
        new_total = current_total + max(1, int(packages)) * base
        self.window.core.config.set('ctx.records.limit.total', new_total)
        self.window.core.config.save()

        # If folders are at top (default), we can safely append without rebuilding.
        # Otherwise, fall back to safe rebuild with stable scroll restore.
        folders_top = bool(self.window.core.config.get("ctx.records.folders.top"))

        # Pull fresh meta with increased total
        data = self.window.core.ctx.get_meta(reload=True)

        if folders_top:
            # compute which ungrouped & not pinned IDs are already visible
            view = self.window.ui.nodes['ctx.list']
            visible_ids = set()
            try:
                visible_ids = view.get_visible_unpaged_ids()
            except Exception:
                pass

            # build candidate list in provider order
            candidates = []
            for mid, meta in data.items():
                if (meta.group_id is None or meta.group_id == 0) and not meta.important:
                    candidates.append(mid)

            # filter only new ones (keep order)
            add_ids = [mid for mid in candidates if mid not in visible_ids]

            # append without rebuilding
            self.window.ui.contexts.ctx_list.append_unpaginated('ctx.list', data, add_ids)
        else:
            # Fallback: rare config where groups are placed after items – do a stable rebuild
            view = self.window.ui.nodes['ctx.list']
            try:
                sb = view.verticalScrollBar()
                prev_val = sb.value()
                # mark special refresh to avoid collisions with helpers
                self._infinite_scroll_refresh = True
                # set pending to preserve old position during rebuild
                view.set_pending_v_scroll(prev_val)
            except Exception:
                prev_val = None

            try:
                self.update_list(reload=True, restore_scroll=True)
            finally:
                self._infinite_scroll_refresh = False

    def _inject_current_if_missing(self, data: dict) -> dict:
        """
        Inject currently selected meta at the end of the list if it's not in the current page.
        This produces a 'placeholder' entry that will be replaced by the real one once loaded.
        """
        try:
            cur_id = self.window.core.ctx.get_current()
            if cur_id is None:
                cur_id = self.window.core.config.get('ctx')
            if not cur_id or cur_id in data:
                return data

            provider = self.window.core.ctx.get_provider()
            meta = provider.get_meta_by_id(cur_id)
            if meta is None:
                return data  # nothing to inject

            # Mark as placeholder in 'extra' for potential future diagnostics; UI ignores this flag.
            try:
                if meta.extra is None:
                    meta.extra = {}
                meta.extra['__placeholder'] = True
            except Exception:
                pass

            injected = dict(data)  # keep original ordering, append at the end
            injected[meta.id] = meta
            return injected
        except Exception:
            return data


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.14 20:00:00                  #
# ================================================== #

class TabsDebug:
    def __init__(self, window=None):
        """
        Tabs debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'tabs'

    def update(self):
        """Update debug window."""
        debug = self.window.core.debug
        tabs_controller = self.window.controller.ui.tabs
        tabs_core = self.window.core.tabs
        ctx_output = self.window.core.ctx.output
        ctx_container = self.window.core.ctx.container

        debug.begin(self.id)
        debug.add(self.id, 'current Col', str(tabs_controller.get_current_column_idx()))
        debug.add(self.id, 'current IDX', str(tabs_controller.get_current_idx()))
        debug.add(self.id, 'current Tab', str(tabs_controller.get_current_tab()))
        debug.add(self.id, 'current PID', str(tabs_controller.get_current_pid()))
        debug.add(self.id, 'current Type', str(tabs_controller.get_current_type()))
        debug.add(self.id, '----', '')
        debug.add(self.id, 'last_pid', str(tabs_core.last_pid))
        debug.add(self.id, 'locked', str(tabs_controller.locked))
        debug.add(self.id, 'col', str(tabs_controller.col))
        debug.add(self.id, 'count(pids)', str(len(tabs_core.pids)))
        debug.add(self.id, 'count(ctx bags)', str(len(ctx_container.bags)))
        debug.add(self.id, '----', '')

        # PIDs
        for pid in list(tabs_core.pids):
            tab = tabs_core.pids[pid]
            data = tab.to_dict()
            debug.add(self.id, f"PID [{pid}]", str(data))

        # mapping PID => meta.id
        debug.add(self.id, '----', '')
        debug.add(self.id, 'PID => meta.id', str(ctx_output.mapping))
        debug.add(self.id, '(last) meta.id => PID', str(ctx_output.last_pids))
        debug.add(self.id, '(last) PID', str(ctx_output.last_pid))
        debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.24 23:00:00                  #
# ================================================== #

from __future__ import annotations
from dataclasses import dataclass, field
from typing import List, Optional, Dict, Any

from agents import TResponseInputItem


@dataclass
class MemoryState:
    """Holds conversation items and last response id for a memory node."""
    mem_id: str
    items: List[TResponseInputItem] = field(default_factory=list)
    last_response_id: Optional[str] = None

    def is_empty(self) -> bool:
        return not self.items

    def set_from(self, items: List[TResponseInputItem], last_response_id: Optional[str]) -> None:
        self.items = list(items or [])
        self.last_response_id = last_response_id

    def update_from_result(self, items: List[TResponseInputItem], last_response_id: Optional[str]) -> None:
        self.set_from(items, last_response_id)


class MemoryManager:
    """Manages MemoryState instances keyed by memory node id."""
    def __init__(self) -> None:
        self._mem: Dict[str, MemoryState] = {}

    def get(self, mem_id: str) -> MemoryState:
        if mem_id not in self._mem:
            self._mem[mem_id] = MemoryState(mem_id=mem_id)
        return self._mem[mem_id]

    def set(self, mem_id: str, items: List[TResponseInputItem], last_response_id: Optional[str]) -> None:
        self.get(mem_id).set_from(items, last_response_id)

    def snapshot(self) -> Dict[str, MemoryState]:
        return dict(self._mem)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.14 20:00:00                  #
# ================================================== #

class AssistantsDebug:
    def __init__(self, window=None):
        """
        Assistants debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'assistants'

    def update(self):
        """Update debug window."""
        debug = self.window.core.debug
        assistant_threads = self.window.controller.assistant.threads
        assistant_editor = self.window.controller.assistant.editor
        assistants_core = self.window.core.assistants
        assistants_store = assistants_core.store

        debug.begin(self.id)
        debug.add(self.id, '(thread) started', str(assistant_threads.started))
        debug.add(self.id, '(thread) stop', str(assistant_threads.stop))
        debug.add(self.id, '(thread) run_id', str(assistant_threads.run_id))
        debug.add(self.id, '(thread) tool_calls', str(assistant_threads.tool_calls))
        debug.add(self.id, 'Options', str(assistant_editor.get_options()))
        debug.add(self.id, '----', '')

        # assistants
        assistants = assistants_core.get_all()
        for key in list(assistants):
            prefix = f"[{key}] "
            assistant = assistants[key]
            data = {
                'id': assistant.id,
                'name': assistant.name,
                'description': assistant.description,
                'model': assistant.model,
                'instructions': assistant.instructions,
                'meta': assistant.meta,
                'tools': assistant.tools,
                'files': assistant.files,
                'tools[function]': assistant.tools['function']
            }
            debug.add(self.id, str(assistant.name), str(data))

        debug.add(self.id, '----', '')

        store_items = {}
        for id in assistants_store.items:
            store_items[id] = assistants_store.items[id].to_dict()

        debug.add(self.id, 'Store (items)', str(store_items))
        debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.02 20:00:00                  #
# ================================================== #

import re
from typing import Optional, List, Dict, Union

from PySide6.QtCore import QTimer
from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import QTextEdit

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_ASSISTANT,
    MODE_CHAT,
    MODE_EXPERT,
    MODE_AGENT_OPENAI,
)
from pygpt_net.controller.presets.editor import Editor
    # Editor controller
from pygpt_net.core.events import AppEvent
from pygpt_net.item.preset import PresetItem
from pygpt_net.utils import trans


_FILENAME_SANITIZE_RE = re.compile(r'[^a-zA-Z0-9_\-\.]')
# keep original validation (do not break other parts)
_VALIDATE_FILENAME_RE = re.compile(r'[^\w\s\-\.]')


class Presets:
    def __init__(self, window=None):
        """
        Presets controller

        :param window: Window instance
        """
        self.window = window
        self.editor = Editor(window)
        self.locked = False
        self.selected = []

    def setup(self):
        """Setup presets"""
        self.editor.setup()

    def on_changed(self):
        """Handle change event"""
        pass

    def is_bot(self) -> bool:
        """
        Check if current preset is bot

        :return: True if bot
        """
        w = self.window
        cfg = w.core.config
        if cfg.get('mode') != MODE_AGENT_OPENAI:
            return False
        preset_id = cfg.get('preset')
        if not preset_id or preset_id == "*":
            return False
        preset_data = w.core.presets.get_by_id(MODE_AGENT_OPENAI, preset_id)
        return bool(
            preset_data
            and preset_data.agent_openai
            and preset_data.agent_provider_openai
            and preset_data.agent_provider_openai.startswith("openai_agent_bot")
        )

    def select(self, idx: int):
        """
        Select preset by list index (legacy)

        :param idx: value of the list (row idx)
        """
        if self.preset_change_locked():
            return
        w = self.window
        mode = w.core.config.get('mode')
        self.set_by_idx(mode, idx)
        preset_id = w.core.config.get('preset')
        w.controller.ui.update()
        w.controller.model.select_current()
        w.dispatch(AppEvent(AppEvent.PRESET_SELECTED))
        self.set_selected(idx)
        editor_ctrl = w.controller.presets.editor
        if editor_ctrl.opened and editor_ctrl.current != preset_id:
            self.editor.init(preset_id)

    def select_by_id(self, preset_id: str):
        """
        Select preset by explicit ID (robust for DnD-ordered views).
        """
        if self.preset_change_locked():
            return
        if not preset_id:
            return
        w = self.window
        mode = w.core.config.get('mode')
        if not w.core.presets.has(mode, preset_id):
            return
        w.core.config.set("preset", preset_id)
        if 'current_preset' not in w.core.config.data:
            w.core.config.data['current_preset'] = {}
        w.core.config.data['current_preset'][mode] = preset_id
        self.select_model()
        w.controller.ui.update()
        w.controller.model.select_current()
        w.dispatch(AppEvent(AppEvent.PRESET_SELECTED))
        idx = w.core.presets.get_idx_by_id(mode, preset_id)
        self.set_selected(idx)
        editor_ctrl = w.controller.presets.editor
        if editor_ctrl.opened and editor_ctrl.current != preset_id:
            self.editor.init(preset_id)

    def get_current(self) -> Optional[PresetItem]:
        """
        Get current preset

        :return: preset item
        """
        w = self.window
        cfg = w.core.config
        preset_id = cfg.get('preset')
        mode = cfg.get('mode')
        if preset_id and w.core.presets.has(mode, preset_id):
            return w.core.presets.get_by_id(mode, preset_id)
        return None

    def next(self):
        """Select next preset"""
        nodes = self.window.ui.nodes
        models = self.window.ui.models
        count = models['preset.presets'].rowCount()
        if count <= 0:
            return
        idx = (nodes['preset.presets'].currentIndex().row() + 1) % count
        self.select(idx)

    def prev(self):
        """Select previous preset"""
        nodes = self.window.ui.nodes
        models = self.window.ui.models
        count = models['preset.presets'].rowCount()
        if count <= 0:
            return
        idx = (nodes['preset.presets'].currentIndex().row() - 1) % count
        self.select(idx)

    def use(self):
        """Copy preset prompt to input"""
        w = self.window
        w.controller.chat.common.append_to_input(
            w.ui.nodes['preset.prompt'].toPlainText()
        )

    def paste_prompt(
            self,
            idx: int,
            parent: str = "global"
    ):
        """
        Paste prompt from template

        :param idx: prompt index
        :param parent: parent name
        """
        w = self.window
        template = w.core.prompt.template.get_by_id(idx)
        if template is None:
            return
        target = None
        if parent == "global":
            target = w.ui.nodes['preset.prompt']
        elif parent == "input":
            target = w.ui.nodes['input']
        elif parent == "editor":
            target = w.ui.config["preset"]["prompt"]
        if target is not None:
            self.paste_to_textarea(target, template['prompt'])

    def paste_custom_prompt(
            self,
            idx: int,
            parent: str = "global"
    ):
        """
        Paste prompt from custom template

        :param idx: prompt index
        :param parent: parent name
        """
        w = self.window
        template = w.core.prompt.custom.get_by_id(idx)
        if template is None:
            return
        target = None
        if parent == "global":
            target = w.ui.nodes['preset.prompt']
        elif parent == "input":
            target = w.ui.nodes['input']
        elif parent == "editor":
            target = w.ui.config["preset"]["prompt"]
        if target is not None:
            self.paste_to_textarea(target, template.content)

    def save_prompt(
            self,
            name: str = "",
            force: bool = False
    ):
        """
        Save prompt to file

        :param name: prompt name
        :param force: force save
        """
        w = self.window
        content = w.ui.nodes['input'].toPlainText()
        if content.strip() == "":
            w.update_status("Prompt is empty!")
            return
        if not force:
            dlg = w.ui.dialog['rename']
            dlg.id = 'prompt.custom.new'
            dlg.input.setText(content[:40] + "...")
            dlg.current = ""
            dlg.show()
            return
        w.ui.dialog['rename'].close()
        w.core.prompt.custom.new(name, content)
        w.update_status("Prompt saved")

    def rename_prompt(
            self,
            uuid: str,
            name: str = "",
            force: bool = False
    ):
        """
        Rename prompt

        :param uuid: prompt ID
        :param name: new name
        :param force: force rename
        """
        w = self.window
        item = w.core.prompt.custom.get_by_id(uuid)
        if item is None:
            return
        if not force:
            dlg = w.ui.dialog['rename']
            dlg.id = 'prompt.custom.rename'
            dlg.input.setText(item.name)
            dlg.current = uuid
            dlg.show()
            return
        w.ui.dialog['rename'].close()
        item.name = name
        w.core.prompt.custom.save()
        w.update_status("Prompt renamed")

    def delete_prompt(
            self,
            uuid: str,
            force: bool = False
    ):
        """
        Delete prompt

        :param uuid: prompt ID
        :param force: force delete
        """
        w = self.window
        item = w.core.prompt.custom.get_by_id(uuid)
        if item is None:
            return
        if not force:
            dlg = w.ui.dialog['confirm']
            dlg.type = 'prompt.custom.delete'
            dlg.id = uuid
            dlg.message = "Are you sure you want to delete this prompt?"
            dlg.show()
            return
        w.ui.dialog['confirm'].close()
        w.core.prompt.custom.delete(uuid)
        w.update_status("Prompt deleted")

    def paste_to_textarea(
            self,
            textarea: QTextEdit,
            text: str
    ):
        """
        Paste text to textarea

        :param textarea: textarea widget
        :param text: text to paste
        """
        prev_text = textarea.toPlainText()
        cur = textarea.textCursor()
        cur.beginEditBlock()
        cur.movePosition(QTextCursor.End)
        txt = str(text).strip()
        if prev_text.strip() != "":
            cur.insertText("\n")
        if txt:
            cur.insertText(txt)
        cur.endEditBlock()
        textarea.setTextCursor(cur)
        textarea.setFocus()

    def set(
            self,
            mode: str,
            preset_id: str
    ):
        """
        Set preset

        :param mode: mode name
        :param preset_id: preset ID
        """
        w = self.window
        if not w.core.presets.has(mode, preset_id):
            return False
        w.core.config.data['preset'] = preset_id
        if 'current_preset' not in w.core.config.data:
            w.core.config.data['current_preset'] = {}
        w.core.config.data['current_preset'][mode] = preset_id

    def set_by_idx(
            self,
            mode: str,
            idx: int
    ):
        """
        Set preset by index

        :param mode: mode name
        :param idx: preset index
        """
        w = self.window
        preset_id = w.core.presets.get_by_idx(idx, mode)
        w.core.config.set("preset", preset_id)
        if 'current_preset' not in w.core.config.data:
            w.core.config.data['current_preset'] = {}
        w.core.config.data['current_preset'][mode] = preset_id
        self.select_model()

    def select_current(self, no_scroll: bool = False):
        """
        Select preset by current

        :param no_scroll: do not scroll to current preset
        """
        w = self.window
        cfg = w.core.config
        mode = cfg.get('mode')
        preset_id = cfg.get('preset')
        if not preset_id:
            return
        idx = w.core.presets.get_idx_by_id(mode, preset_id)
        if idx is None or idx < 0:
            return
        if no_scroll:
            w.ui.nodes['preset.presets'].store_scroll_position()
        w.ui.nodes['preset.presets'].select_by_idx(idx)
        if no_scroll:
            w.ui.nodes['preset.presets'].restore_scroll_position()
        editor_ctrl = w.controller.presets.editor
        if editor_ctrl.opened and editor_ctrl.current != preset_id:
            self.editor.init(preset_id)

    def select_default(self):
        """Set default preset"""
        w = self.window
        cfg = w.core.config
        preset_id = cfg.get('preset')
        if not preset_id:
            mode = cfg.get('mode')
            current = cfg.get('current_preset')
            if mode in current and current[mode] and current[mode] in w.core.presets.get_by_mode(mode):
                cfg.set('preset', current[mode])
            else:
                cfg.set('preset', w.core.presets.get_default(mode))
            new_id = cfg.get('preset')
            editor_ctrl = w.controller.presets.editor
            if editor_ctrl.opened and editor_ctrl.current != new_id:
                self.editor.init(new_id)

    def update_data(self):
        """Update preset data"""
        w = self.window
        cfg = w.core.config
        preset_id = cfg.get('preset')
        if not preset_id:
            self.reset()
            w.controller.mode.reset_current()
            return
        if preset_id not in w.core.presets.items:
            cfg.set('preset', "")
            self.reset()
            w.controller.mode.reset_current()
            return
        preset = w.core.presets.items[preset_id]
        w.ui.nodes['preset.prompt'].setPlainText(preset.prompt)
        w.core.config.set('prompt', preset.prompt)
        w.core.config.set('ai_name', preset.ai_name)
        w.core.config.set('user_name', preset.user_name)
        w.core.config.set('agent.llama.provider', preset.agent_provider)
        w.core.config.set('agent.openai.provider', preset.agent_provider_openai)
        w.core.config.set('agent.llama.idx', preset.idx)

    def update_current(self):
        """Update current mode, model and preset"""
        w = self.window
        cfg = w.core.config
        mode = cfg.get('mode')
        preset_id = cfg.get('preset')
        if preset_id and preset_id in w.core.presets.items:
            preset = w.core.presets.items[preset_id]
            cfg.set('user_name', preset.user_name)
            cfg.set('ai_name', preset.ai_name)
            cfg.set('prompt', preset.prompt)
            cfg.set('temperature', preset.temperature)
            cfg.set('agent.llama.provider', preset.agent_provider)
            cfg.set('agent.openai.provider', preset.agent_provider_openai)
            cfg.set('agent.llama.idx', preset.idx)
            return
        cfg.set('user_name', None)
        cfg.set('ai_name', None)
        cfg.set('temperature', 1.0)
        if mode == MODE_CHAT:
            cfg.set('prompt', w.core.prompt.get('default'))
        else:
            cfg.set('prompt', None)

    def get_current_functions(self) -> Optional[List[Dict]]:
        """
        Get current preset functions

        :return: list of functions
        """
        w = self.window
        preset_id = w.core.config.get('preset')
        if preset_id and preset_id in w.core.presets.items:
            preset = w.core.presets.items[preset_id]
            if preset.has_functions():
                return preset.get_functions()
        return None

    def from_global(self):
        """Update current preset from global prompt"""
        if self.locked:
            return
        w = self.window
        preset_id = w.core.config.get('preset')
        if preset_id and preset_id in w.core.presets.items:
            preset = w.core.presets.items[preset_id]
            preset.prompt = w.core.config.get('prompt')
            w.core.presets.save(preset_id)

    def select_model(self):
        """Select model by current preset"""
        w = self.window
        cfg = w.core.config
        mode = cfg.get('mode')
        preset_id = cfg.get('preset')
        if not preset_id or preset_id not in w.core.presets.items:
            return
        preset = w.core.presets.items[preset_id]
        model = preset.model
        if not model or model == "_":
            return
        models = w.core.models
        if models.has(model) and models.is_allowed(model, mode):
            if cfg.get('model') == model:
                return
            cfg.set('model', model)
            w.controller.model.set(mode, model)
            w.controller.model.init_list()
            w.controller.model.select_current()

    def refresh(self, no_scroll: bool = False):
        """
        Refresh presets

        :param no_scroll: do not scroll to current
        """
        self.locked = True
        w = self.window
        if w.core.config.get('mode') == MODE_ASSISTANT:
            self.locked = False
            return
        if no_scroll:
            w.ui.nodes['preset.presets'].store_scroll_position()
        self.select_default()
        self.update_current()
        self.update_data()
        w.controller.mode.update_temperature()
        self.update_list()
        self.select_current()
        if no_scroll:
            w.ui.nodes['preset.presets'].restore_scroll_position()
        self.on_changed()
        self.locked = False

    def update_list(self):
        """Update presets list"""
        w = self.window
        mode = w.core.config.get('mode')
        items = w.core.presets.get_by_mode(mode)
        w.ui.toolbox.presets.update_presets(items)
        self.clear_selected()

    def reset(self):
        """Reset preset data"""
        self.window.ui.nodes['preset.prompt'].setPlainText("")

    def make_filename(self, name: str) -> str:
        """
        Make preset filename from name

        :param name: preset name
        :return: preset filename
        """
        return _FILENAME_SANITIZE_RE.sub('_', name.lower())

    def duplicate(self, idx: Optional[Union[int, list]] = None):
        """
        Duplicate preset

        :param idx: preset ID (or list of IDs)
        """
        ids = idx if isinstance(idx, list) else [idx]
        last_id = None
        updated = False
        w = self.window
        mode = w.core.config.get('mode')
        for idx in ids:
            if idx is not None:
                if len(ids) > 1:
                    preset = w.core.presets.get_by_id(mode, idx)  # by ID, PresetItem
                    preset_id = preset.filename if preset else None
                else:
                    preset = w.core.presets.get_by_idx(idx, mode) # by index, str
                    preset_id = preset
                if preset_id:
                    new_id = w.core.presets.duplicate(preset_id)
                    last_id = new_id
                    updated = True

        if updated:
            self.update_list()
            self.refresh(no_scroll=True)
            if len(ids) == 1 and idx is not None:
                idx = w.core.presets.get_idx_by_id(mode, last_id)
                self.editor.edit(idx)
                self.select(idx)  # switch to the new preset if only one was duplicated
            w.update_status(trans('status.preset.duplicated'))


    def enable(self, idx: Optional[int] = None):
        """
        Enable preset

        :param idx: preset index (row index)
        """
        if idx is not None:
            w = self.window
            mode = w.core.config.get('mode')
            preset_id = w.core.presets.get_by_idx(idx, mode)
            if preset_id and preset_id in w.core.presets.items:
                w.core.presets.enable(preset_id)
                QTimer.singleShot(100, self.refresh)

    def disable(self, idx: Optional[int] = None):
        """
        Disable preset

        :param idx: preset index (row index)
        """
        if idx is not None:
            w = self.window
            mode = w.core.config.get('mode')
            preset_id = w.core.presets.get_by_idx(idx, mode)
            if preset_id and preset_id in w.core.presets.items:
                w.core.presets.disable(preset_id)
                QTimer.singleShot(100, self.refresh)

    def clear(self, force: bool = False):
        """
        Clear preset data

        :param force: force clear data
        """
        w = self.window
        preset = w.core.config.get('preset')
        if not force:
            w.ui.dialogs.confirm(
                type='preset_clear',
                id='',
                msg=trans('confirm.preset.clear'),
            )
            return
        w.core.config.set('prompt', "")
        w.core.config.set('ai_name', "")
        w.core.config.set('user_name', "")
        w.core.config.set('temperature', 1.0)
        if preset and preset in w.core.presets.items:
            p = w.core.presets.items[preset]
            p.ai_name = ""
            p.user_name = ""
            p.prompt = ""
            p.temperature = 1.0
            self.refresh()
        w.update_status(trans('status.preset.cleared'))
        mode = w.core.config.get('mode')
        if mode == MODE_ASSISTANT:
            w.core.assistants.load()
            w.core.remote_store.openai.load_all()

    def _nearest_id_after_delete(self, mode: str, idx: Optional[int], deleting_id: Optional[str]) -> Optional[str]:
        """
        Compute the nearest neighbor to select after deletion:
        - Prefer the next item (below) if exists;
        - Otherwise choose the previous one (above);
        - Returns None when no neighbor can be determined.
        This uses the current view order for the given mode, including pinned current.<mode> at index 0.
        """
        try:
            w = self.window
            data = w.core.presets.get_by_mode(mode) or {}
            ids = list(data.keys())
            if not ids:
                return None

            # If idx is invalid, try to resolve from deleting_id
            if idx is None or idx < 0 or idx >= len(ids):
                if deleting_id and deleting_id in ids:
                    idx = ids.index(deleting_id)
                else:
                    return None

            # Prefer below
            if idx + 1 < len(ids):
                cand = ids[idx + 1]
                if cand and cand != deleting_id:
                    return cand

            # Otherwise above
            if idx - 1 >= 0:
                cand = ids[idx - 1]
                if cand and cand != deleting_id:
                    return cand
        except Exception:
            pass
        return None

    def delete(
            self,
            idx: Optional[Union[int, list]] = None,
            force: bool = False
    ):
        """
        Delete preset

        :param idx: preset index (row index) or list of indices
        :param force: force delete without confirmation
        """
        ids = idx if isinstance(idx, list) else [idx]
        w = self.window
        mode = w.core.config.get('mode')
        is_preset = False
        updated = False
        for tmp_id in ids:
            if len(ids) > 1:
                preset = w.core.presets.get_by_id(mode, tmp_id)  # by ID, PresetItem
            else:
                preset = w.core.presets.get_by_idx(tmp_id, mode)  # by index, str
            if preset:
                is_preset = True
                break

        if ids and is_preset:
            if not force:
                w.ui.dialogs.confirm(
                    type='preset_delete',
                    id=idx,
                    msg=trans('confirm.preset.delete'),
                )
                return

            # Determine neighbor only if the deleted preset is currently active.
            # This keeps API semantics untouched and prevents unexpected selection changes.
            for idx in ids:
                if len(ids) > 1:
                    preset = w.core.presets.get_by_id(mode, idx) # by ID, PresetItem
                    preset_id = preset.filename if preset else None
                else:
                    preset = w.core.presets.get_by_idx(idx, mode) # by index, str
                    preset_id = preset
                if not preset:
                    continue
                is_current = (preset_id == w.core.config.get('preset'))
                target_id = None
                if is_current:
                    target_id = self._nearest_id_after_delete(mode, idx, preset_id)

                # Remove from core (also removes file when True)
                w.core.presets.remove(preset_id, True)

                # When removing the active preset, jump to the nearest neighbor (below or above).
                # If no neighbor can be determined, fall back to previous behavior (clear and let defaults apply).
                if is_current:
                    if target_id and target_id in w.core.presets.items:
                        # Persist new selection in config (and keep current_preset mapping coherent)
                        w.core.config.set('preset', target_id)
                        if 'current_preset' not in w.core.config.data:
                            w.core.config.data['current_preset'] = {}
                        w.core.config.data['current_preset'][mode] = target_id
                    else:
                        # Fallback: clear selection to allow select_default() to pick the first available
                        w.core.config.set('preset', None)
                        w.ui.nodes['preset.prompt'].setPlainText("")

                updated = True

        if updated:
            self.refresh(no_scroll=True)
            w.update_status(trans('status.preset.deleted'))

    def restore(self, force: bool = False):
        """
        Restore preset data

        :param force: force restore data
        """
        w = self.window
        if not force:
            w.ui.dialogs.confirm(
                type='preset_restore',
                id='',
                msg=trans('confirm.preset.restore'),
            )
            return
        mode = w.core.config.get('mode')
        if mode == MODE_AGENT:
            mode = MODE_EXPERT
        w.core.presets.restore(mode)
        self.refresh()

    def is_current(self, idx: Optional[int] = None) -> bool:
        """
        Check if preset is current

        :param idx: preset index (row index)
        :return: True if current
        """
        if idx is not None:
            w = self.window
            mode = w.core.config.get('mode')
            if mode == MODE_AGENT:
                mode = MODE_EXPERT
            preset_id = w.core.presets.get_by_idx(idx, mode)
            if preset_id:
                if preset_id == "current." + mode:
                    return True
        return False

    def validate_filename(self, value: str) -> str:
        """
        Validate filename

        :param value: filename
        :return: sanitized filename
        """
        return _VALIDATE_FILENAME_RE.sub('', value)

    def preset_change_locked(self) -> bool:
        """
        Check if preset change is locked

        :return: True if locked
        """
        return self.window.core.config.get('mode') == MODE_ASSISTANT

    def reload(self):
        """Reload presets"""
        was_locked = self.locked
        self.locked = True
        self.window.core.presets.load()
        self.refresh()
        if not was_locked:
            self.locked = False

    def lock(self):
        """Lock presets change"""
        self.locked = True

    def unlock(self):
        """Unlock presets change"""
        self.locked = False

    def add_selected(self, id: int):
        """
        Add selection ID to selected list

        :param id: preset ID
        """
        if id not in self.selected:
            self.selected.append(id)

    def remove_selected(self, id: int):
        """
        Remove selection ID from selected list

        :param id: preset ID
        """
        if id in self.selected:
            self.selected.remove(id)

    def set_selected(self, id: int):
        """
        Set selected ID in selected list

        :param id: preset ID
        """
        self.selected = [id] if id is not None else []
        self.on_changed()

    def clear_selected(self):
        """Clear selected list"""
        self.selected = []

    # ----------------------------
    # Drag & drop ordering helpers
    # ----------------------------

    def persist_order_for_mode(self, mode: str, uuids: List[str]):
        """
        Persist new order (by UUIDs) for given mode.

        The special '*' preset (current.<mode>) is not included here and always pinned at index 0.
        """
        w = self.window
        cfg = w.core.config
        order = cfg.get('presets_order') or {}
        # Normalize to lists
        if isinstance(order.get(mode), dict):
            mapped = order.get(mode)
            order[mode] = [mapped[k] for k in sorted(mapped.keys(), key=lambda x: int(x))]
        order[mode] = [u for u in uuids if u]
        cfg.set('presets_order', order)

    def dnd_enabled(self) -> bool:
        """
        Check if drag and drop is globally enabled in config.
        """
        try:
            return bool(self.window.core.config.get('presets.drag_and_drop.enabled'))
        except Exception:
            return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.15 22:00:00                  #
# ================================================== #

from PySide6.QtCore import QItemSelectionModel
from PySide6.QtWidgets import QTreeView, QAbstractItemView


class BaseList(QTreeView):
    NAME = range(1)  # list of columns

    def __init__(self, window=None, id=None):
        """
        Select menu

        :param window: Window instance
        :param id: input id
        """
        super(BaseList, self).__init__(window)
        self.window = window
        self.id = id
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.setIndentation(0)
        self.selection_locked = None
        self.selection = None
        self.unlocked = False
        self.clicked.connect(self.click)
        self.header().hide()
        self.v_scroll_value = 0
        self.h_scroll_value = 0

        # pending scroll values applied while updates are disabled (to avoid top flicker)
        self._pending_v_scroll_value = None
        self._pending_h_scroll_value = None

    def click(self, val):
        self.window.controller.mode.select(self.id)
        self.selection = self.selectionModel().selection()

    def lockSelection(self, selected=None, deselected=None):
        if self.selection is not None:
            self.selectionModel().select(self.selection, QItemSelectionModel.Select)

    def backup_selection(self):
        self.selection = self.selectionModel().selection()

    def restore_selection(self):
        if self.selection is not None:
            self.selectionModel().select(self.selection, QItemSelectionModel.Select)

    def mousePressEvent(self, event):
        index = self.indexAt(event.pos())
        if not index.isValid():
            return
        super(BaseList, self).mousePressEvent(event)

    def focusOutEvent(self, event):
        pass

    def selectionCommand(self, index, event=None):
        """
        Selection command
        :param index: Index
        :param event: Event
        """
        # check tmp unlock
        if self.unlocked:
            return super().selectionCommand(index, event)
        if self.selection_locked is not None and self.selection_locked():
            return QItemSelectionModel.NoUpdate
        return super().selectionCommand(index, event)

    def select_by_idx(self, idx: int):
        """
        Select item by index

        :param idx: index
        """
        if idx < 0:
            return
        model = self.model()
        if model.rowCount() > idx:
            index = model.index(idx, 0)
            prev_unlocked = self.unlocked
            self.unlocked = True
            self.selectionModel().select(
                index, QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows
            )
            self.setCurrentIndex(index)
            self.setFocus()
            self.scrollTo(index)
            self.unlocked = prev_unlocked

    def store_scroll_position(self):
        """Store current scroll position"""
        self.v_scroll_value = self.verticalScrollBar().value()
        self.h_scroll_value = self.horizontalScrollBar().value()

    def restore_scroll_position(self):
        """Restore scroll position"""
        self.verticalScrollBar().setValue(self.v_scroll_value)
        self.horizontalScrollBar().setValue(self.h_scroll_value)

    def set_pending_v_scroll(self, value: int):
        """
        Set vertical scroll value to apply while updates are disabled.
        This prevents a visible jump to the top during model rebuild.
        """
        self._pending_v_scroll_value = int(value)

    def set_pending_h_scroll(self, value: int):
        """Optional: set horizontal pending value."""
        self._pending_h_scroll_value = int(value)

    def clear_pending_scroll(self):
        """Clear pending scroll values."""
        self._pending_v_scroll_value = None
        self._pending_h_scroll_value = None

    def apply_pending_scroll(self):
        """
        Apply pending scroll values immediately.
        IMPORTANT: Call this before re-enabling updates to avoid repaint at top.
        """
        if self._pending_v_scroll_value is not None:
            self.verticalScrollBar().setValue(self._pending_v_scroll_value)
        if self._pending_h_scroll_value is not None:
            self.horizontalScrollBar().setValue(self._pending_h_scroll_value)
        # do not clear here; let caller decide when to clear



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.16 22:00:00                  #
# ================================================== #

from typing import Optional, List, Dict

from pygpt_net.item.ctx import CtxMeta
from pygpt_net.core.tabs.tab import Tab


class Output:

    def __init__(self, window=None):
        """
        Context output mapping

        :param window: Window
        """
        self.window = window
        # [column_idx -> {pid -> meta_id}]
        self.mapping: Dict[int, Dict[int, int]] = {}
        # [column_idx -> {meta_id -> pid}] (last used PID for meta)
        self.last_pids: Dict[int, Dict[int, int]] = {}
        self.last_pid: int = 0  # last used PID
        self.initialized: bool = False

    def init(self, force: bool = False):
        """
        Initialize mappings

        :param force: Force reinitialize
        """
        if self.initialized and not force:
            return

        self.mapping.clear()
        self.last_pids.clear()

        tabs = getattr(self.window, "core", None)
        tabs = getattr(tabs, "tabs", None)
        num_cols = getattr(tabs, "NUM_COLS", 0) if tabs is not None else 0

        for n in range(num_cols):
            self.mapping[n] = {}
            self.last_pids[n] = {}

        self.initialized = True

    def store(self, meta: CtxMeta) -> int:
        """
        Store meta in mapping

        :param meta: Meta
        :return: PID
        """
        self.init()

        tabs = self.window.core.tabs
        pid = tabs.get_active_pid()
        tab = tabs.get_tab_by_pid(pid)
        if tab is None or tab.type != Tab.TAB_CHAT:
            return 0

        col_idx = tab.column_idx

        col_map = self.mapping.get(col_idx)
        if col_map is None or not isinstance(col_map, dict):
            col_map = self.mapping[col_idx] = {}
        last_map = self.last_pids.get(col_idx)
        if last_map is None:
            last_map = self.last_pids[col_idx] = {}

        col_map[pid] = meta.id
        last_map[meta.id] = pid
        self.last_pid = pid

        tab.data_id = meta.id
        return pid

    def is_mapped(self, meta: CtxMeta) -> bool:
        """
        Check if meta is mapped anywhere

        :param meta: Meta
        :return: True if mapped, False otherwise
        """
        self.init()

        # quick check for the active tab
        tabs = self.window.core.tabs
        active_pid = tabs.get_active_pid()
        tab = tabs.get_tab_by_pid(active_pid)
        if tab is not None:
            col_map = self.mapping.get(tab.column_idx, {})
            # check if meta.id is in the current column mapping
            for mid in col_map.values():
                if mid == meta.id:
                    return True

        # fallback
        for col_idx, col_map in self.mapping.items():
            if tab is not None and col_idx == tab.column_idx:
                continue
            for mid in col_map.values():
                if mid == meta.id:
                    return True
        return False

    def get_meta(self, pid: int) -> Optional[int]:
        """
        Get meta by PID

        :param pid: PID
        :return: meta ID or None
        """
        self.init()

        tabs = self.window.core.tabs
        tab = tabs.get_tab_by_pid(pid)
        if tab is not None:
            return self.mapping.get(tab.column_idx, {}).get(pid)

        # fallback
        for col_map in self.mapping.values():
            if pid in col_map:
                return col_map[pid]
        return None

    def prepare_meta(self, tab: Tab) -> Optional[int]:
        """
        Get meta ID by PID (and store it if not exists)

        :param tab: Tab
        :return: Meta ID or None
        """
        self.init()

        pid = tab.pid
        col_idx = tab.column_idx

        col_map = self.mapping.get(col_idx)
        if col_map is None:
            col_map = self.mapping[col_idx] = {}

        if pid in col_map:
            return col_map[pid]

        meta_id = getattr(tab, "data_id", None)
        if meta_id is not None:
            col_map[pid] = meta_id
            self.last_pids.setdefault(col_idx, {})[meta_id] = pid
            self.last_pid = pid
        return meta_id

    def get_mapped(self, meta: CtxMeta) -> Optional[int]:
        """
        Get PID by meta for the current column (prefer active PID)

        :param meta: Meta
        :return: PID or None
        """
        self.init()

        tabs = self.window.core.tabs
        active_pid = tabs.get_active_pid()
        tab = tabs.get_tab_by_pid(active_pid)
        if tab is None:
            return None
        col_idx = tab.column_idx
        col_map = self.mapping.get(col_idx, {})
        if not isinstance(col_map, dict):
            return None

        candidates = [pid for pid, meta_id in col_map.items() if meta_id == meta.id]
        for pid in candidates:
            if pid == active_pid:  # prefer active PID
                return pid
        return candidates[-1] if candidates else None  # return last found PID

    def is_empty(self) -> bool:
        """
        Check if mapping is empty for the active PID across columns

        :return: True if empty, False otherwise
        """
        self.init()

        tabs = self.window.core.tabs
        active_pid = tabs.get_active_pid()
        for col_map in self.mapping.values():
            if active_pid in col_map:
                return False
        return True

    def get_pid(self, meta: Optional[CtxMeta] = None) -> Optional[int]:
        """
        Get PID by meta (prefer active PID)

        :param meta: Meta
        :return: PID or None
        """
        self.init()
        if meta is None:
            return None

        tabs = self.window.core.tabs
        active_pid = tabs.get_active_pid()
        mapped_pid = self.get_mapped(meta)
        if mapped_pid == active_pid:
            return active_pid
        return self.store(meta)

    def clear(self):
        """Clear mapping"""
        self.mapping.clear()
        self.last_pids.clear()
        self.last_pid = 0
        self.initialized = False

    def get_current(self, meta: Optional[CtxMeta] = None):
        """
        Get current output node by meta

        :param meta: Meta
        :return: Node
        """
        pid = self.get_pid(meta)
        nodes = self.window.ui.nodes['output']
        if pid is not None:
            node = nodes.get(pid)
            if node is not None:
                return node
        # fallback
        return next(iter(nodes.values()), None)

    def get_current_plain(self, meta: Optional[CtxMeta] = None):
        """
        Get current output plain node by meta

        :param meta: Meta
        :return: Node
        """
        pid = self.get_pid(meta)
        nodes = self.window.ui.nodes['output_plain']
        if pid is not None:
            node = nodes.get(pid)
            if node is not None:
                return node
        return next(iter(nodes.values()), None)

    def get_by_pid(self, pid: Optional[int] = None):
        """
        Get output node by PID

        :param pid: PID
        :return: Node widget
        """
        nodes = self.window.ui.nodes['output']
        if pid is not None:
            node = nodes.get(pid)
            if node is not None:
                return node
        return next(iter(nodes.values()), None)

    def get_by_pid_plain(self, pid: Optional[int] = None):
        """
        Get output plain node by PID

        :param pid: PID
        :return: Node widget
        """
        nodes = self.window.ui.nodes['output_plain']
        if pid is not None:
            node = nodes.get(pid)
            if node is not None:
                return node
        return next(iter(nodes.values()), None)

    def get_all(self) -> List:
        """
        Get all output nodes

        :return: List of nodes
        """
        return list(self.window.ui.nodes['output'].values())

    def get_all_plain(self) -> List:
        """
        Get all output plain nodes

        :return: List of nodes
        """
        return list(self.window.ui.nodes['output_plain'].values())

    def remove_pid(self, pid: int):
        """
        Remove PID from mapping

        :param pid: PID
        """
        self.init()
        for col_idx in self.mapping:
            if pid in self.mapping[col_idx]:
                del self.mapping[col_idx][pid]
                break
        if pid in self.last_pids:
            del self.last_pids[pid]
        if pid == self.last_pid:
            self.last_pid = 0


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.15 23:00:00                  #
# ================================================== #

from typing import Any, Dict, Optional, Union


class Input:
    def __init__(self, window=None):
        """
        Text input field handler

        :param window: Window instance
        """
        self.window = window

    def apply(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ):
        """
        Apply value to input

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        """
        if "value" not in option:
            return

        ui = self.window.ui
        parent = ui.config.get(parent_id)
        if not parent:
            return
        widget = parent.get(key)
        if widget is None:
            return

        value = option["value"]
        typ = option.get("type")

        if typ == "int":
            try:
                value = round(int(value), 0)
            except (ValueError, TypeError):
                value = 0
        elif typ == "float":
            try:
                value = float(value)
            except (ValueError, TypeError):
                value = 0.0

        if typ in ("int", "float") and value is not None:
            minv = option.get("min")
            maxv = option.get("max")
            if minv is not None and value < minv:
                value = minv
            elif maxv is not None and value > maxv:
                value = maxv

        new_text = str(value)
        curr_val = None
        if hasattr(widget, 'text'):
            curr_val = widget.text()
        elif hasattr(widget, 'toPlainText'):
            curr_val = widget.toPlainText()
        if curr_val != new_text:
            widget.setText(new_text)

    def on_update(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any],
            value: Any,
            hooks: bool = True,
            only_hook: bool = False,
    ):
        """
        On update event handler

        :param parent_id: Parent ID
        :param key: Option key
        :param option: Option data
        :param value: Option value
        :param hooks: Run hooks
        :param only_hook: Only run hook, do not apply value
        """
        option['value'] = value
        if not only_hook:
            self.apply(parent_id, key, option)

        if hooks:
            ui = self.window.ui
            hook_name = f"update.{parent_id}.{key}"
            if ui.has_hook(hook_name):
                hook = ui.get_hook(hook_name)
                try:
                    hook(key, value, "input")
                except Exception as e:
                    self.window.core.debug.log(e)

    def get_value(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ) -> Optional[Union[int, float, str]]:
        """
        Get value from input

        :param parent_id: Parent ID
        :param key: Option key
        :param option: Option data
        :return: Value
        """
        typ = option["type"]
        widget = self.window.ui.config[parent_id][key]
        text = widget.text()

        if typ == "int":
            try:
                return int(text)
            except (ValueError, TypeError):
                return 0
        elif typ == "float":
            try:
                return float(text)
            except (ValueError, TypeError):
                return 0.0
        elif typ == "text":
            return text
        return None


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.15 00:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        # Endpoints / HTTP
        plugin.add_option(
            "api_base",
            type="text",
            value="https://api.github.com",
            label="API base",
            description="Base API URL (default https://api.github.com).",
        )
        plugin.add_option(
            "web_base",
            type="text",
            value="https://github.com",
            label="Web base",
            description="GitHub website base (default https://github.com).",
        )
        plugin.add_option(
            "api_version",
            type="text",
            value="2022-11-28",
            label="API version header",
            description="X-GitHub-Api-Version value.",
        )
        plugin.add_option(
            "http_timeout",
            type="int",
            value=30,
            label="HTTP timeout (s)",
            description="Requests timeout in seconds.",
        )

        # OAuth Device Flow
        plugin.add_option(
            "oauth_client_id",
            type="text",
            value="",
            label="OAuth Client ID",
            description="Client ID from your GitHub OAuth App (supports Device Flow).",
            secret=True,
        )
        plugin.add_option(
            "oauth_scopes",
            type="text",
            value="repo read:org read:user user:email",
            label="Scopes",
            description="Space-separated OAuth scopes for Device Flow.",
        )
        plugin.add_option(
            "oauth_open_browser",
            type="bool",
            value=True,
            label="Open browser automatically",
            description="Open verification URL in default browser.",
        )
        plugin.add_option(
            "oauth_auto_begin",
            type="bool",
            value=True,
            label="Auto-start auth when required",
            description="Begin Device Flow automatically when a command needs a token.",
        )
        plugin.add_option(
            "oauth_scope_granted",
            type="text",
            value="",
            label="(auto) Granted scopes",
            description="Set after authorization.",
        )

        # Tokens
        plugin.add_option(
            "gh_access_token",
            type="textarea",
            value="",
            label="(auto) OAuth access token",
            description="Stored OAuth access token (Device/Web).",
            secret=True,
        )
        plugin.add_option(
            "pat_token",
            type="textarea",
            value="",
            label="PAT token (optional)",
            description="Personal Access Token (classic or fine-grained).",
            secret=True,
        )
        plugin.add_option(
            "auth_scheme",
            type="text",
            value="Bearer",
            label="Auth scheme",
            description="Bearer or token (PAT usually 'token').",
        )

        # Convenience cache
        plugin.add_option(
            "user_id",
            type="text",
            value="",
            label="(auto) User ID",
            description="Cached after gh_me or auth.",
        )
        plugin.add_option(
            "username",
            type="text",
            value="",
            label="(auto) Username",
            description="Cached after gh_me or auth.",
        )

        # ---------------- Commands ----------------

        # Auth
        plugin.add_cmd(
            "gh_device_begin",
            instruction="Begin OAuth Device Flow (returns user_code and verification URL).",
            params=[
                {"name": "scopes", "type": "str", "required": False, "description": "Override scopes (space-separated)"},
            ],
            enabled=True,
            description="Auth: device begin",
            tab="auth",
        )
        plugin.add_cmd(
            "gh_device_poll",
            instruction="Poll device code for access token.",
            params=[
                {"name": "device_code", "type": "str", "required": True, "description": "Device code from gh_device_begin"},
            ],
            enabled=True,
            description="Auth: device poll",
            tab="auth",
        )
        plugin.add_cmd(
            "gh_set_pat",
            instruction="Set Personal Access Token (PAT).",
            params=[
                {"name": "token", "type": "str", "required": True, "description": "PAT token value"},
                {"name": "scheme", "type": "str", "required": False, "description": "Auth scheme: token|Bearer"},
            ],
            enabled=True,
            description="Auth: set PAT",
            tab="auth",
        )

        # Users
        plugin.add_cmd(
            "gh_me",
            instruction="Get authenticated user.",
            params=[],
            enabled=True,
            description="Users: me",
            tab="users",
        )
        plugin.add_cmd(
            "gh_user_get",
            instruction="Get user by username.",
            params=[{"name": "username", "type": "str", "required": True, "description": "GitHub login"}],
            enabled=True,
            description="Users: get",
            tab="users",
        )

        # Repos
        plugin.add_cmd(
            "gh_repos_list",
            instruction="List repositories.",
            params=[
                {"name": "username", "type": "str", "required": False, "description": "List for user (public)"},
                {"name": "org", "type": "str", "required": False, "description": "List for organization"},
                {"name": "type", "type": "str", "required": False, "description": "all|owner|member (for /user/repos)"},
                {"name": "visibility", "type": "str", "required": False, "description": "all|public|private"},
                {"name": "sort", "type": "str", "required": False, "description": "created|updated|pushed|full_name"},
                {"name": "direction", "type": "str", "required": False, "description": "asc|desc"},
                {"name": "per_page", "type": "int", "required": False, "description": "Items per page"},
                {"name": "page", "type": "int", "required": False, "description": "Page index"},
            ],
            enabled=True,
            description="Repos: list",
            tab="repos",
        )
        plugin.add_cmd(
            "gh_repo_get",
            instruction="Get repository details.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner login"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo name"},
            ],
            enabled=True,
            description="Repos: get",
            tab="repos",
        )
        plugin.add_cmd(
            "gh_repo_create",
            instruction="Create repository for user or org.",
            params=[
                {"name": "name", "type": "str", "required": True, "description": "Repository name"},
                {"name": "description", "type": "str", "required": False, "description": "Description"},
                {"name": "private", "type": "bool", "required": False, "description": "Default False"},
                {"name": "auto_init", "type": "bool", "required": False, "description": "Default True"},
                {"name": "org", "type": "str", "required": False, "description": "Organization (if creating in org)"},
            ],
            enabled=True,
            description="Repos: create",
            tab="repos",
        )
        plugin.add_cmd(
            "gh_repo_delete",
            instruction="Delete repository (dangerous).",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner login"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo name"},
                {"name": "confirm", "type": "bool", "required": False, "description": "Must be true"},
            ],
            enabled=False,
            description="Repos: delete",
            tab="repos",
        )

        # Contents
        plugin.add_cmd(
            "gh_contents_get",
            instruction="Get file or list directory contents.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner login"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo name"},
                {"name": "path", "type": "str", "required": False, "description": "Path inside repo (optional)"},
                {"name": "ref", "type": "str", "required": False, "description": "Branch/tag/commit SHA"},
            ],
            enabled=True,
            description="Contents: get",
            tab="contents",
        )
        plugin.add_cmd(
            "gh_file_put",
            instruction="Create or update a file via Contents API.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner login"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo name"},
                {"name": "path", "type": "str", "required": True, "description": "Path inside repo"},
                {"name": "message", "type": "str", "required": False, "description": "Commit message"},
                {"name": "content", "type": "str", "required": False, "description": "Raw text content"},
                {"name": "local_path", "type": "str", "required": False, "description": "Local file to upload"},
                {"name": "branch", "type": "str", "required": False, "description": "Branch name"},
                {"name": "sha", "type": "str", "required": False, "description": "Existing blob sha for update"},
                {"name": "committer", "type": "dict", "required": False, "description": '{"name":"","email":""}'},
                {"name": "author", "type": "dict", "required": False, "description": '{"name":"","email":""}'},
                {"name": "resolve_sha", "type": "bool", "required": False, "description": "Auto-detect sha if missing"},
            ],
            enabled=True,
            description="Contents: put file",
            tab="contents",
        )
        plugin.add_cmd(
            "gh_file_delete",
            instruction="Delete a file via Contents API.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner login"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo name"},
                {"name": "path", "type": "str", "required": True, "description": "Path inside repo"},
                {"name": "message", "type": "str", "required": True, "description": "Commit message"},
                {"name": "sha", "type": "str", "required": True, "description": "Blob sha"},
                {"name": "branch", "type": "str", "required": False, "description": "Branch name"},
            ],
            enabled=True,
            description="Contents: delete file",
            tab="contents",
        )

        # Issues
        plugin.add_cmd(
            "gh_issues_list",
            instruction="List repository issues.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo"},
                {"name": "state", "type": "str", "required": False, "description": "open|closed|all"},
                {"name": "labels", "type": "str", "required": False, "description": "Comma labels"},
                {"name": "creator", "type": "str", "required": False, "description": "Creator login"},
                {"name": "mentioned", "type": "str", "required": False, "description": "Mentioned login"},
                {"name": "assignee", "type": "str", "required": False, "description": "Assignee login"},
                {"name": "since", "type": "str", "required": False, "description": "ISO8601"},
                {"name": "per_page", "type": "int", "required": False, "description": "Items per page"},
                {"name": "page", "type": "int", "required": False, "description": "Page"},
            ],
            enabled=True,
            description="Issues: list",
            tab="issues",
        )
        plugin.add_cmd(
            "gh_issue_create",
            instruction="Create an issue.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo"},
                {"name": "title", "type": "str", "required": True, "description": "Issue title"},
                {"name": "body", "type": "str", "required": False, "description": "Issue body"},
                {"name": "assignees", "type": "list", "required": False, "description": "List of logins"},
                {"name": "labels", "type": "list", "required": False, "description": "List of labels"},
            ],
            enabled=True,
            description="Issues: create",
            tab="issues",
        )
        plugin.add_cmd(
            "gh_issue_comment",
            instruction="Comment on an issue.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo"},
                {"name": "number", "type": "int", "required": True, "description": "Issue number"},
                {"name": "body", "type": "str", "required": True, "description": "Comment body"},
            ],
            enabled=True,
            description="Issues: comment",
            tab="issues",
        )
        plugin.add_cmd(
            "gh_issue_close",
            instruction="Close an issue.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo"},
                {"name": "number", "type": "int", "required": True, "description": "Issue number"},
            ],
            enabled=True,
            description="Issues: close",
            tab="issues",
        )

        # Pull Requests
        plugin.add_cmd(
            "gh_pulls_list",
            instruction="List pull requests.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo"},
                {"name": "state", "type": "str", "required": False, "description": "open|closed|all"},
                {"name": "head", "type": "str", "required": False, "description": "user:branch"},
                {"name": "base", "type": "str", "required": False, "description": "Base branch"},
                {"name": "sort", "type": "str", "required": False, "description": "created|updated|popularity|long-running"},
                {"name": "direction", "type": "str", "required": False, "description": "asc|desc"},
                {"name": "per_page", "type": "int", "required": False, "description": "Items per page"},
                {"name": "page", "type": "int", "required": False, "description": "Page"},
            ],
            enabled=True,
            description="PR: list",
            tab="pulls",
        )
        plugin.add_cmd(
            "gh_pull_create",
            instruction="Create a pull request.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo"},
                {"name": "title", "type": "str", "required": True, "description": "PR title"},
                {"name": "head", "type": "str", "required": True, "description": "Source branch (user:branch)"},
                {"name": "base", "type": "str", "required": True, "description": "Target branch"},
                {"name": "body", "type": "str", "required": False, "description": "PR body"},
                {"name": "draft", "type": "bool", "required": False, "description": "Draft PR"},
            ],
            enabled=True,
            description="PR: create",
            tab="pulls",
        )
        plugin.add_cmd(
            "gh_pull_merge",
            instruction="Merge a pull request.",
            params=[
                {"name": "owner", "type": "str", "required": True, "description": "Owner"},
                {"name": "repo", "type": "str", "required": True, "description": "Repo"},
                {"name": "number", "type": "int", "required": True, "description": "PR number"},
                {"name": "commit_title", "type": "str", "required": False, "description": "Commit title"},
                {"name": "commit_message", "type": "str", "required": False, "description": "Commit message"},
                {"name": "merge_method", "type": "str", "required": False, "description": "merge|squash|rebase"},
            ],
            enabled=True,
            description="PR: merge",
            tab="pulls",
        )

        # Search
        plugin.add_cmd(
            "gh_search_repos",
            instruction="Search repositories.",
            params=[
                {"name": "q", "type": "str", "required": True, "description": "Query string"},
                {"name": "sort", "type": "str", "required": False, "description": "stars|forks|help-wanted-issues|updated"},
                {"name": "order", "type": "str", "required": False, "description": "desc|asc"},
                {"name": "per_page", "type": "int", "required": False, "description": "Items per page"},
                {"name": "page", "type": "int", "required": False, "description": "Page"},
            ],
            enabled=True,
            description="Search: repositories",
            tab="search",
        )
        plugin.add_cmd(
            "gh_search_issues",
            instruction="Search issues and pull requests.",
            params=[
                {"name": "q", "type": "str", "required": True, "description": "Query string"},
                {"name": "sort", "type": "str", "required": False, "description": "comments|reactions|created|updated"},
                {"name": "order", "type": "str", "required": False, "description": "desc|asc"},
                {"name": "per_page", "type": "int", "required": False, "description": "Items per page"},
                {"name": "page", "type": "int", "required": False, "description": "Page"},
            ],
            enabled=True,
            description="Search: issues",
            tab="search",
        )
        plugin.add_cmd(
            "gh_search_code",
            instruction="Search code.",
            params=[
                {"name": "q", "type": "str", "required": True, "description": "Query string (use qualifiers)"},
                {"name": "per_page", "type": "int", "required": False, "description": "Items per page"},
                {"name": "page", "type": "int", "required": False, "description": "Page"},
            ],
            enabled=True,
            description="Search: code",
            tab="search",
        )